from reho.paths import *
import pandas as pd
import numpy as np
import datetime

__doc__ = """
Characterizes the CO2 emissions related to electricity generated from the grid.
"""


def find_average_value(country, metric, emissions_matrix):
    # sort city to country
    emissions_matrix.columns = np.arange(1, 8761)
    emissions_matrix = emissions_matrix.xs((country, metric), level=(0, 2))

    average = emissions_matrix.mean(axis=1)
    if (metric == 'GWP100a') or (metric == 'GWP20a'):
        average = average / 1000  # g/kWh to kg/kWh
    return average


def annual_to_typical_emissions(cluster, country, metric, df_time, df_emission):
    # get relevant cluster information
    if isinstance(cluster, pd.DataFrame):
        PeriodDuration = cluster[["TimeEnd"]]
    else:
        PeriodDuration = pd.DataFrame(np.concatenate([np.repeat(cluster['PeriodDuration'], cluster['Periods']), [1, 1]]), columns=["TimeEnd"])
        PeriodDuration.index = PeriodDuration.index + 1
    df_emission.columns = np.arange(1, 8761)

    # construct Multiindex
    df_p = pd.DataFrame()
    list_timesteps = []

    for p in df_time.index:
        start = datetime.datetime(2005, 1, 1)
        date = df_time.xs(p).Date
        difference = date - start
        dif_h = int(difference.total_seconds() / 3600)

        end = PeriodDuration.xs(p + 1).TimeEnd  # ampl starts at 1
        start = dif_h + 1
        ende = dif_h + end
        df_emission_p = df_emission.xs((country, metric), level=(0, 2))

        df_emission_p = df_emission_p.loc[:, start: ende]
        np_period = df_emission_p.values[0]
        if (metric == 'GWP100a') or (metric == 'GWP20a'):
            np_period = np_period / 1000  # g/kWh to kg/kWh

        df_period = pd.DataFrame(np_period)
        for t in np.arange(1, int(end) + 1):  # ampl starts at 1
            list_timesteps.append(('Electricity', p + 1, t))  # create ampl index

        df_p = pd.concat((df_p, df_period))

    idx = pd.MultiIndex.from_tuples(list_timesteps)

    # marry index and data
    df_E = df_p.set_index(idx)
    df_E = df_E.rename(columns={0: 'GWP_supply'})

    return df_E


def return_typical_emission_profiles(cluster, File_ID, metric, timestamp_file, emissions_matrix):
    country = 'CH'
    emission_file = os.path.join(path_to_clustering, metric + '_' + File_ID + '.dat')

    if os.path.exists(emission_file):
        df_E = pd.read_csv(emission_file, index_col=[0, 1, 2])
    else:
        df_E = annual_to_typical_emissions(cluster, country, metric, timestamp_file, emissions_matrix)
        df_E.to_csv(emission_file)

    return df_E



import configparser
import csv
import math
import os.path
import re

import geopandas as gpd
import numpy as np
import pandas as pd
from sqlalchemy import create_engine, MetaData, select
from sqlalchemy.dialects import postgresql

import reho.model.preprocessing.skydome as skydome
from reho.paths import *


class QBuildingsReader:
    """
    This class is used to handle and prepare the data related to buildings.

    There usually come from `GBuildings <https://ipese-web.epfl.ch/lepour/qbuildings/index.html>`_ database. However,
    one can use data from a csv, in which case the column names should correspond to the GBuildings one, described in
    `Processed GBuildings tables <https://ipese-web.epfl.ch/lepour/qbuildings/GBuildings/description.html#processed>`_.

    Parameters
    ----------
    load_facades : bool
        Whether the facades data should be added.
    load_roofs : bool
        Whether the roofs data should be added.
    """

    def __init__(self, load_facades=False, load_roofs=False):

        self.db = None
        self.tables = None
        self.db_schema = None
        self.db_engine = None
        self.connection = None
        self.data = {}
        self.load_facades = load_facades
        self.load_roofs = load_roofs

        self.local_data = dict()
        self.local_data["df_Area"] = pd.read_csv(path_to_areas, header=None)
        self.local_data["df_Cenpts"] = pd.read_csv(path_to_cenpts, header=None)

    def establish_connection(self, db):
        """
        Allows to establish the connection with one of the QBuildings database.

        Parameters
        ----------
        db : str
            Name of the database to which we want to connect

        Notes
        -----
        - It is highly recommend to pick 'Suisse' as a database as it is the only one maintained at the moment.
        """
        # Database connection
        file_ini = path_to_qbuildings + "/" + db + ".ini"

        project = configparser.ConfigParser()
        project.read(file_ini)

        # Database
        self.db_schema = "Processed"
        try:
            db_engine_str = 'postgresql+psycopg2://{}:{}@{}:{}/{}'.format(project['database']['username'],
                                                                          project['database']['password'],
                                                                          project['database']['host'],
                                                                          project['database']['port'],
                                                                          project['database']['database'])
            self.db_engine = create_engine(db_engine_str)
            self.connection = self.db_engine.connect()  # test connection
            print('Connected to database')

        except Exception as e:
            print(f'Cannot connect to database engine: {e}')

        if 'database' in project:
            print('\thost: {}\n\tport: {}\n\tdatabase: {}\n\tusername: {}'.format(
                project['database']['host'],
                project['database']['port'],
                project['database']['database'],
                project['database']['username']))

        # input
        metadata = MetaData(bind=self.db_engine)
        metadata.reflect(schema=self.db_schema)
        self.tables = metadata.tables
        self.db = db

        return

    def read_csv(self, buildings_filename='buildings.csv', nb_buildings=None, roofs_filename='roofs.csv', facades_filename='facades.csv'):
        """
        Reads buildings-related data from CSV files and prepare it for the REHO model.

        If not all the buildings from the file should be extracted, one can give a number of buildings.
        The fields from the files are translated to the corresponding ones used in REHO.

        Parameters
        ----------
        buildings_filename : str
            The filename of the CSV file containing buildings data.
        nb_buildings : int, optional
            The number of buildings to consider. If not provided, all buildings in the file are considered.
        roofs_filename : str, optional
            The filename of the CSV file containing roofs data.
        facades_filename : str, optional
            The filename of the CSV file containing facades data.

        Returns
        -------
        dict
            A dictionary containing the prepared data for the REHO model, including buildings, facades, roofs,
            and shadows if roofs and facades are loaded.

        Notes
        -----
        - If `nb_buildings` is not provided, all buildings in the 'buildings' data are considered.
        - If ``load_roofs = True``, `roofs_filename` must be provided, else it is not useful. Same goes for the facades.
        - This function can be used with default files in case one does not want to connect to the database and does
          not need a particular building.
          In that case, do not fill any filename. `buildings.csv`, `roofs.csv` and `facades.csv`
          will be used by default.
          It should be noted that those names are therefore reserved for the default and cannot be used for your own files.

        Example
        -------
        >>> from reho.model.reho import *
        >>> reader = QBuildingsReader(load_roofs=True)
        >>> qbuildings_data = reader.read_csv("buildings.csv", roofs_filename="roofs.csv", nb_buildings=7)

        >>> qbuildings_data['buildings_data'].keys()
        dict_keys(['Building1', 'Building2', 'Building3'])

        >>> qbuildings_data['buildings_data']['Building1'].keys()
        dict_keys(['id_class', 'ratio', 'status', 'ERA', 'SolarRoofArea', 'area_facade_m2', 'height_m', 'U_h', 'HeatCapacity', 'T_comfort_min_0', 'Th_supply_0', 'Th_return_0', 'Tc_supply_0', 'Tc_return_0', 'x', 'y', 'z', 'geometry', 'transformer', 'id_building', 'egid', 'period', 'n_p', 'energy_heating_signature_kWh_y', 'energy_cooling_signature_kWh_y', 'energy_hotwater_signature_kWh_y', 'energy_el_kWh_y'])
        """
        self.data['buildings'] = file_reader(buildings_filename)
        self.data['buildings'] = translate_buildings_to_REHO(self.data['buildings'])
        # self.data['buildings'] = add_geometry(self.data['buildings'])

        if nb_buildings is None:
            nb_buildings = self.data['buildings'].shape[0]
        buildings = self.select_buildings_data(nb_buildings, None)
        # buildings = add_geometry(buildings)
        qbuildings = {'buildings_data': buildings}
        if self.load_facades:
            self.data['facades'] = file_reader(path_handler(facades_filename))
            selected_facades = self.select_roofs_or_facades_data(roof=False)
            self.data['facades'] = self.data['facades'][self.data['facades'].index.isin(selected_facades)]
            self.data['facades'] = add_geometry(self.data['facades'])
            self.data['facades'] = translate_facades_to_REHO(self.data['facades'], self.data['buildings'])
            qbuildings['facades_data'] = self.data['facades']
            qbuildings['shadows_data'] = return_shadows_district(qbuildings['buildings_data'], self.data['facades'], self.local_data)

        if self.load_roofs:
            self.data['roofs'] = file_reader(path_handler(roofs_filename))
            selected_roofs = self.select_roofs_or_facades_data(roof=True)
            self.data['roofs'] = self.data['roofs'][self.data['roofs'].index.isin(selected_roofs)]
            self.data['roofs'] = add_geometry(self.data['roofs'])
            self.data['roofs'] = translate_roofs_to_REHO(self.data['roofs'])
            qbuildings['roofs_data'] = self.data['roofs']

        return qbuildings

    def read_db(self, transformer=None, nb_buildings=None, egid=None, to_csv=False):
        """
        Reads the database and extract from it the buildings required, by the LV transformer's ID.

        If not all the buildings from the transformer should be extracted, one can give a number of buildings or if
        the EGIDs are known, pass a list of EGIDs.
        The fields from the database are translated to the corresponding ones used in REHO.

        Parameters
        ----------
        transformer : int
            ID of the transformer on which we want to optimize
        nb_buildings : int
            Number of buildings to select
        egid : list
            To specify a list of buildings to optimize with their EGIDs
        to_csv : bool
            To export the data into csv

        Returns
        -------
        dict
            A dictionary that contains the qbuildings data. The default has only one key ``buildings_data``
            with a dictionary of buildings, with their fields and corresponding values.


        Notes
        -----
        - The use of this function requires the previous creation of a ``QBuildingsReader`` and the use of ``establish_connection('Suisse')``.
        - EGIDs are the postal address unique identifier used in Switzerland. One can find the EGIDs of a given address at the `RegBL <https://www.housing-stat.ch/fr/query/adrtoegid.html>`_.
        - If ``load_roofs = True`` the roofs are returned as well in the dictionary as a DataFrame under the key ``roofs_data``.
        - If ``load_facades = True`` the facades and the shadows are returned as well in the dictionary as a DataFrame under the keys ``roofs_data`` and ``shadows_data``.

        Examples
        --------
        >>> from reho.model.reho import *
        >>> reader = QBuildingsReader(load_roofs=True)
        >>> reader.establish_connection('Suisse')
        >>> qbuildings_data = reader.read_db(transformer=3658, egid=[954117])

        >>> qbuildings_data['buildings_data']
        {'buildings_data': {'Building1': {'id_class': 'I', 'ratio': '1.0', 'status': "['existing', 'existing', 'existing']", 'ERA': 1396.0, 'SolarRoofArea': 1121.8206745917826, 'area_facade_m2': 848.6771960464813, 'height_m': 9.211343577064236, 'U_h': 0.00152, 'HeatCapacity': 120.29999999999991, 'T_comfort_min_0': 20.0, 'Th_supply_0': 65.0, 'Th_return_0': 50.0, 'Tc_supply_0': 12.0, 'Tc_return_0': 17.0, 'x': 2592703.9673297284, 'y': 1120087.7339999992, 'z': 572.4461527539248, 'geometry': <POLYGON ((2592684.383 1120074.623, 2592683.644 1120075.443, 2592679.083 112...>, 'transformer': 3658, 'id_building': '40214', 'egid': '954117', 'period': '1981-1990', 'n_p': 34.9, 'energy_heating_signature_kWh_y': 111855.52745599969, 'energy_cooling_signature_kWh_y': 0.0, 'energy_hotwater_signature_kWh_y': 4562.903646729638, 'energy_el_kWh_y': 39088.0}}

        >>> qbuildings_data['roofs_data']
            TILT  ...                                           geometry
        0     26  ...  MULTIPOLYGON (((2592819.164 1120187.216, 25928...
        1     25  ...  MULTIPOLYGON (((2592832.585 1120154.503, 25928...
        2     25  ...  MULTIPOLYGON (((2592819.164 1120187.216, 25928...
        3     26  ...  MULTIPOLYGON (((2592824.929 1120157.956, 25928...
        0     19  ...  MULTIPOLYGON (((2592378.668 1120324.589, 25923...
        ..   ...  ...                                                ...
        25     0  ...  MULTIPOLYGON (((2592872.699 1120127.178, 25928...
        26     0  ...  MULTIPOLYGON (((2592917.016 1120132.965, 25929...
        27    28  ...  MULTIPOLYGON (((2592891.248 1120129.691, 25928...
        28    26  ...  MULTIPOLYGON (((2592901.604 1120125.591, 25929...
        29    27  ...  MULTIPOLYGON (((2592887.725 1120119.181, 25928...
        [252 rows x 6 columns]

        """

        # TODO: SQL query to select only buildings, roofs and facades of interest

        # Select the right transformer
        sqlQuery = select([self.tables[self.db_schema + '.' + 'transformers']]) \
            .where(self.tables[self.db_schema + '.' + 'transformers'].columns.id == transformer)
        self.data['transformers'] = gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()), con=self.db_engine,
                                                     geom_col='geometry').fillna(np.nan)

        # Select buildings
        sqlQuery = select([self.tables[self.db_schema + '.' + 'buildings']]) \
            .where(self.tables[self.db_schema + '.' + 'buildings'].columns.transformer == transformer)
        self.data['buildings'] = gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()), con=self.db_engine,
                                                  geom_col='geometry').fillna(np.nan)
        mask = (self.data['buildings']['egid'].isnull())
        self.data['buildings'] = self.data['buildings'].loc[~mask, :]

        if nb_buildings is None:
            nb_buildings = self.data['buildings'].shape[0]
        if to_csv:
            self.data['buildings'].to_csv('buildings.csv', index=False)

        self.data['buildings'] = translate_buildings_to_REHO(self.data['buildings'])
        buildings = self.select_buildings_data(nb_buildings, egid)
        if to_csv:
            csv_columns = list(buildings[list(buildings.keys())[0]].keys())
            with open('reho_input.csv', 'w') as csvfile:
                writer = csv.DictWriter(csvfile, csv_columns)
                writer.writeheader()
                for building in buildings:
                    writer.writerow(buildings[building])

        qbuildings = {'buildings_data': buildings}

        if self.load_facades:
            # TODO: Correct the roofs and facades selection with the id filtered by select_buildings_data
            self.data['facades'] = gpd.GeoDataFrame()
            for id in self.data['buildings'].id_building:
                sqlQuery = select([self.tables[self.db_schema + '.' + 'facades']]) \
                    .where(self.tables[self.db_schema + '.' + 'facades'].columns.id_building == id)
                self.data['facades'] = pd.concat(
                    (self.data['facades'], gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()),
                                                            con=self.db_engine, geom_col='geometry').fillna(np.nan)))
            if to_csv:
                self.data['facades'].to_csv('facades.csv', index=False)
            self.data['facades'] = translate_facades_to_REHO(self.data['facades'], self.data['buildings'])
            qbuildings['facades_data'] = self.data['facades']
            qbuildings['shadows_data'] = return_shadows_district(qbuildings["buildings_data"], self.data['facades'], self.local_data)
        if self.load_roofs:
            self.data['roofs'] = gpd.GeoDataFrame()
            for id in self.data['buildings'].id_building:
                sqlQuery = select([self.tables[self.db_schema + '.' + 'roofs']]) \
                    .where(self.tables[self.db_schema + '.' + 'roofs'].columns.id_building == id)
                self.data['roofs'] = pd.concat(
                    (self.data['roofs'], gpd.read_postgis(sqlQuery.compile(dialect=postgresql.dialect()),
                                                          con=self.db_engine, geom_col='geometry').fillna(np.nan)))
            if to_csv:
                self.data['roofs'].to_csv('roofs.csv', index=False)
            self.data['roofs'] = translate_roofs_to_REHO(self.data['roofs'])
            qbuildings['roofs_data'] = self.data['roofs']

        if qbuildings["buildings_data"] == {}:
            raise print("Empty building data")

        return qbuildings

    def select_buildings_data(self, nb_buildings, egid=None):

        if egid is None:
            nb_select = 0
            selected_buildings = []
            reindex = []
            for i, building in self.data['buildings'].iterrows():
                # Only execute optimization for complete dictionary else skip and count
                if re.search('XIII', building['id_class']) is None:
                    selected_buildings.append(building['id_building'])
                    nb_select += 1
                    reindex.append("Building" + str(nb_select))
                if nb_select >= nb_buildings:
                    break
            self.data['buildings'] = self.data['buildings'][
                self.data['buildings']['id_building'].isin(selected_buildings)]
            self.data['buildings'].index = reindex
            if self.db_engine is None:
                self.data['buildings'] = add_geometry(self.data['buildings'])
            buildings_data = self.data['buildings'].to_dict('index')

        else:
            nb_select = 1
            if type(egid) != list:
                egid = [egid]

            buildings_data = gpd.GeoDataFrame()
            for i in egid:
                data_single_bui = self.data['buildings'][self.data['buildings']['egid'] == str(i)]
                data_single_bui.index = ["Building" + str(nb_select)]
                nb_select += 1
                buildings_data = pd.concat([buildings_data, data_single_bui])
            buildings_data = buildings_data.to_dict('index')

        return buildings_data

    def select_roofs_or_facades_data(self, roof):
        selected_data = []
        for i, building in self.data['buildings'].iterrows():
            if roof:
                selected_data += \
                    self.data['roofs'].index[self.data['roofs']['id_building'] == building['id_building']].to_list()
            else:
                selected_data += \
                    self.data['facades'].index[self.data['facades']['id_building'] == building['id_building']].to_list()

        return selected_data


def translate_buildings_to_REHO(df_buildings):
    new_buildings_data = gpd.GeoDataFrame()
    dict_QBuildings_REHO = {

        #################################################
        # Data strictly necessary for a REHO optimization
        #################################################

        # Data for EUD profiles
        'id_building': 'id_building',
        'egid': 'egid',
        'id_class': 'id_class',
        'class': 'class',
        'ratio': 'ratio',
        'status': 'status',
        'period': 'period',
        'capita_cap': 'n_p',
        # 'NOMSECTEUR': 'NOMSECTEUR',

        # Area
        'area_era_m2': 'ERA',
        'area_roof_solar_m2': 'SolarRoofArea',
        'area_facade_m2': 'area_facade_m2',
        'height_m': 'height_m',  # only for use facades
        'count_floor': 'count_floor',

        # Heating source
        'source_heating': 'source_heating',
        'source_hotwater': 'source_hotwater',

        # Thermal envelope
        'thermal_transmittance_signature_kW_m2_K': 'U_h',
        'thermal_specific_capacity_Wh_m2_K': 'HeatCapacity',

        # Temperature requirements
        'temperature_interior_C': 'T_comfort_min_0',
        'temperature_heating_supply_C': 'Th_supply_0',
        'temperature_heating_return_C': 'Th_return_0',
        'temperature_cooling_supply_C': 'Tc_supply_0',
        'temperature_cooling_return_C': 'Tc_return_0',

        #############################
        # Data not strictly necessary
        #############################

        # Geographic information
        'x': 'x',
        'y': 'y',
        'z': 'z',
        'geometry': 'geometry',
        'transformer': 'transformer',

        # Annual energy
        'energy_heating_signature_kWh_y': 'energy_heating_signature_kWh_y',
        'energy_cooling_signature_kWh_y': 'energy_cooling_signature_kWh_y',
        'energy_hotwater_signature_kWh_y': 'energy_hotwater_signature_kWh_y',
        'energy_el_kWh_y': 'energy_el_kWh_y',

        # Annual roof and facade irradiance
        'roof_annual_irr_kWh_y': 'roof_annual_irr_kWh_y',
        'facade_annual_irr_kWh_y': 'facade_annual_irr_kWh_y'
    }

    # TODO: correct heat source dictionary
    # dict_translate_heat_source = {
    #     'Gas': 'gas',
    #     'Oil': 'oil',
    #     'Wood (beech)': 'wood',
    #     'Electricity': 'electricity',
    #     'Heat pump': 'Heat pump',
    #     'District heat': 'District heat',
    #     'Other': 'gas',
    #     'Other/Oil': 'oil',
    #     'Electricity/Other': 'electricity',
    #     'Oil/Gas': 'oil',
    #     'Oil/Other': 'oil',
    #     'Electricity/Oil': 'electricity',
    #     'Heat pump/Electricity': 'Heat pump',
    #     'Solar collector': 'solar',
    #     'Solar (thermal)': 'solar',
    #     'No energy source/Oil': 'oil',
    #     'No energy source': 'gas',
    #     'Electricity/No energy source': 'electricity',
    #     'Oil/No energy source': 'oil',
    #     'Gas/Solar collector': 'gas',
    #     'Not determined': 'unknown'
    # }

    for key in dict_QBuildings_REHO.keys():
        REHO_index = dict_QBuildings_REHO[key]
        try:
            new_buildings_data[REHO_index] = df_buildings[key]
        except KeyError:
            print('Key %s not in the dictionary' % key)

    # for i, building in new_buildings_data.iterrows():
    #     try:
    #         sources_translated = dict_translate_heat_source[building['source_heating']]
    #         sources_translated_w = dict_translate_heat_source[building['source_hotwater']]
    #         building['source_heating'] = sources_translated
    #         building['source_hotwater'] = sources_translated_w
    #     except:
    #         sources = building['source_heating'].split('/')
    #         try:
    #             sources_translated = dict_translate_heat_source[sources[0]]
    #         except KeyError:
    #             print('Source %s not in the dictionary' % sources[0])
    #             sources_translated = 'not documented'
    #         for source in sources[1:]:
    #             try:
    #                 sources_translated += '/' + dict_translate_heat_source[source]
    #             except KeyError:
    #                 print('Source %s not in the dictionary' % source)
    #                 continue
    #         building['source_heating'] = sources_translated
    #         sources_w = building['source_hotwater'].split('/')
    #         try:
    #             sources_translated_w = dict_translate_heat_source[sources_w[0]]
    #         except KeyError:
    #             print('Source %s not in the dictionary' % sources_w[0])
    #             sources_translated_w = 'not documented'
    #         for source_w in sources_w[1:]:
    #             try:
    #                 sources_translated_w += '/' + dict_translate_heat_source[source_w]
    #             except KeyError:
    #                 print('Source %s not in the dictionary' % source_w)
    #         building['source_hotwater'] = sources_translated_w
    #     new_buildings_data.loc[i, :] = building

    df_buildings = new_buildings_data

    return df_buildings


def translate_facades_to_REHO(df_facades, df_buildings):
    new_facades_data = gpd.GeoDataFrame()
    dict_facades = {'azimuth': 'AZIMUTH',
                    'id_facade': 'Facades_ID',
                    'area_facade_solar_m2': 'AREA',
                    'id_building': 'id_building',
                    # 'cx': 'CX',
                    # 'cy': 'CY',
                    'geometry': 'geometry'}

    for key in dict_facades.keys():
        REHO_index = dict_facades[key]
        try:
            new_facades_data[REHO_index] = df_facades[key]
        except KeyError:
            print('Key %s not in the dictionary' % key)
            
    df_facades = new_facades_data
    df_facades['CX'] = df_facades['geometry'].centroid.x
    df_facades['CY'] = df_facades['geometry'].centroid.y
    df_facades['coord_Z0'] = None

    for i, b in df_buildings.iterrows():
        concerned_facades = df_facades[df_facades['id_building'] == b['id_building']]
        if not pd.isna(b['z']):
            df_facades.loc[concerned_facades.index, 'coord_Z0'] = b['z']

    return df_facades


def translate_roofs_to_REHO(df_roofs):
    # TODO: in Luise fct there is azimuth + 180
    new_roofs_data = gpd.GeoDataFrame()
    dict_roofs = {'tilt': 'TILT',
                  'azimuth': 'AZIMUTH',
                  'id_roof': 'ROOF_ID',
                  'area_roof_solar_m2': 'AREA',
                  'id_building': 'id_building',
                  'geometry': 'geometry'}

    for key in dict_roofs.keys():
        REHO_index = dict_roofs[key]
        try:
            new_roofs_data[REHO_index] = df_roofs[key]
        except KeyError:
            print('Key %s not in the dictionary' % key)
        
    df_roofs = new_roofs_data

    return df_roofs


def get_roofs(self, buildings):
    selected_roofs = []
    for i, building in buildings.iterrows():
        selected_roofs += \
            self.data['roofs'].index[self.data['roofs']['id_building'] == building['id_building']].to_list()
    self.data['roofs'] = self.data['roofs'][self.data['roofs'].index.isin(selected_roofs)]

    return self.data['roofs']


def get_facades(self, buildings):
    selected_facades = []
    for i, building in buildings.iterrows():
        selected_facades += \
            self.data['facades'].index[self.data['facades']['id_building'] == building['id_building']].to_list()
    self.data['facades'] = self.data['facades'][self.data['facades'].index.isin(selected_facades)]

    return self.data['facades']


def calculate_id_building_shadows(df_angles, id_building, local_data):
    df_angles['to_id_building'] = pd.to_numeric(df_angles['to_id_building'])
    df_angles = df_angles.set_index('to_id_building')
    df_angles = df_angles.xs(id_building)

    df_dome = skydome.skydome_to_df(local_data)

    df_shadow = pd.DataFrame()

    for az in df_dome.azimuth.unique():
        df_angles.loc[:, 'cosa2'] = df_angles.apply(lambda x: skydome.f_cos([x['azimuth'], az]), axis=1)
        # df_cosa2 = df_id_building.apply(lambda x: skydome.f_cos([x['azimuth'], az]), axis=1)
        df = df_angles.loc[(df_angles['cosa2'] > 0)].copy()
        # filter buildings which are more than 180 degree apart from patch with az
        df.loc[:, 'tanba'] = df.tanb * df.cosa2
        # calculate tan(beta) for all buildings. Assumption: dxy is the shortest distance and buildings infinite wide

        max_tanba = df['tanba'].max()  # get max obscurance tan(beta, alpha)
        max_b = math.degrees(math.atan(max_tanba))  # get angle

        if math.isnan(max_b):  # if no shadow, set everything to 0
            id_building = 0
            max_b = 0
            max_tanba = 0
        else:
            id_building = df[df.tanba == df['tanba'].max()].index[0]  # get the id_building which causes the obscurance

        df = pd.DataFrame([[max_tanba, max_b, az, id_building]], columns=['tanb', 'beta', 'azimuth', 'id_building'])
        df_shadow = pd.concat((df_shadow, df))

    return df_shadow


def neighbourhood_angles(buildings, facades):
    df_angles = pd.DataFrame()

    for b in buildings:
        id_building = buildings[b]['id_building']
        print('Calculate angles for building: ' + str(id_building))
        df_BUI = pd.DataFrame()
        df_district = {buildings: bui for buildings, bui in buildings.items() if bui['id_building'] != id_building}
        df_district = pd.DataFrame.from_dict(df_district, orient='index')
        df_district = df_district.set_index('id_building')
        # exclude current building to avoid division with zero
        facades_build = facades[facades.id_building == id_building]  # facades of building
        for f in facades_build.index:
            df_c = pd.DataFrame(index=df_district.index)  # df for calculating values for each facade
            df_c['dx'] = df_district.x.values - facades_build.loc[f]['CX']
            df_c['dy'] = df_district.y.values - facades_build.loc[f]['CY']
            df_c['dxy'] = (df_c.dx * df_c.dx + df_c.dy * df_c.dy) ** 0.5
            heights = df_district.height_m.values
            if facades_build.loc[f]['coord_Z0'] is None:
                print('Missing value coord_Z0, not possible to use_facades')
                continue
            df_c['dz'] = df_district.z.values + heights - facades_build.loc[f]['coord_Z0']
            # facades.loc[f]['HEIGHT_Z'] + facades.loc[f]['HEIGHT'] #take foot of facades/ HEIGHT_Z is upperbound
            df_c['tanb'] = df_c.dz / df_c.dxy
            df_c['cosa'] = df_c.dy / df_c.dxy
            df_c['azimuth'] = df_c[['dx', 'dy']].apply(skydome.f_atan, axis=1)
            df_c['beta'] = df_c[['dz', 'dxy']].apply(skydome.f_atan, axis=1)
            df_c = pd.concat([df_c], keys=[f], names=['UID'])
            df_BUI = pd.concat((df_BUI, df_c))

        df_BUI = df_BUI.reset_index()
        df_BUI = df_BUI.rename(columns={'id_building': 'to_id_building'})
        df_BUI['id_building'] = int(id_building)

        df_angles = pd.concat((df_angles, df_BUI))

    df_angles.to_csv('data/angles.csv')

    return df_angles


def return_shadows_district(buildings, facades, local_data):
    df_shadows = pd.DataFrame()

    if os.path.exists('data/angles.csv'):
        df_angles = pd.read_csv('data/angles.csv')
    else:
        df_angles = neighbourhood_angles(buildings, facades)

    for b in buildings:
        id_building = int(buildings[b]['id_building'])
        if id_building in df_angles['id_building'].values:  # check if angle calculation for id_building exists
            df_id_building = calculate_id_building_shadows(df_angles, id_building, local_data)
            idx = np.repeat(id_building, len(df_id_building))
            df_id_building = df_id_building.set_index(idx)
        else:
            print('NO DATA AVAILABLE FOR id_building ' + str(id_building))
            df_id_building = pd.DataFrame(index=[id_building],
                                          columns=['tanb', 'beta', 'azimuth', 'id_building'])  # pass NaN instead
        df_shadows = pd.concat((df_shadows, df_id_building))

    df_shadows["id_building"] = df_shadows["id_building"].astype(str)
    df_shadows.to_csv('data/shadows.csv')

    return df_shadows


def return_shadows_id_building(id_building, df_district, local_data):
    id_building = int(id_building)

    if os.path.isfile('data/shadows.csv'):
        df = file_reader('data/shadows.csv', index_col=0)
    else:
        df = df_district
    df = df.xs(id_building)
    df_dome = skydome.skydome_to_df(local_data)

    df_beta_dome = pd.DataFrame()
    for az in df_dome.azimuth:
        df_beta_dome = pd.concat((df_beta_dome, df.beta[df.azimuth == az]), ignore_index=True)

    df_beta_dome = df_beta_dome.rename(columns={0: 'Limiting_angle_shadow'})

    return df_beta_dome


def add_geometry(df):
    # Avoid issues with geometry when reading data from a csv
    try:
        geom = gpd.GeoSeries.from_wkb(df['geometry'])
    except KeyError:
        print("No geometry in the dataframe")
        return df
    except:
        try:
            geom = gpd.GeoSeries.from_wkt(df['geometry'])
        except TypeError:
            print("Geometry passed is neither of format wkb or wkt so neither from PostGIS, neither from QBuildings")
            return df

    return gpd.GeoDataFrame(df, geometry=geom)



import math
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import calendar
import datetime as dt
from reho.paths import *
from reho.model.preprocessing.clustering import ClusterClass
import pvlib
from pyproj import Transformer


__doc__ = """
Generates the meteorological data (temperature and solar irradiance).
"""


def get_cluster_file_ID(cluster):
    """
    Gets the weather file ID that corresponds to the specifications provided in the reho initalization.

    The file ID is built by concatenating Location_Periods_PeriodDuration_Attributes.
    ``cluster = {'Location': 'Geneva', 'Attributes': ['T', 'I', 'W'], 'Periods': 10, 'PeriodDuration': 24}``
    Will yield to:
    ``File_ID = 'Geneva_10_24_T_I_W'``

    Parameters
    ----------
    cluster : dict
        Dictionary that contains a 'Location' (str), some 'Attributes' (list, among 'T' (temperature), 'I' (irradiance), 'W' (weekday) and 'E' (emissions)),
        a number of periods 'Periods' (int) and a 'PeriodDuration' (int)

    Returns
    -------
    A string that is the file ID.
    """
    if 'T' in cluster['Attributes']:
        T = '_T'
    else:
        T = ''
    if 'I' in cluster['Attributes']:
        I = '_I'
    else:
        I = ''
    if 'W' in cluster['Attributes']:
        W = '_W'
    else:
        W = ''
    if 'E' in cluster['Attributes']:
        E = '_E'
    else:
        E = ''

    File_ID = cluster['Location'] + '_' + str(cluster['Periods']) + '_' + str(cluster['PeriodDuration']) + T + I + W + E

    return File_ID


def generate_weather_data(cluster, qbuildings_data):
    """
    This function is called if the clustered weather data specified by File_ID do not exist yet.
    Runs the ClusterClass and create the required files.
    """

    if 'custom_weather' in cluster.keys():
        df = read_custom_weather(cluster['custom_weather'])
    else:
        df = get_weather_data(qbuildings_data).reset_index(drop=True)

    attributes = []
    if 'T' in cluster['Attributes']:
        attributes.append('Text')
    if 'I' in cluster['Attributes']:
        attributes.append('Irr')
    if 'W' in cluster['Attributes']:
        attributes.append('Weekday')
    if 'E' in cluster['Attributes']:
        attributes.append('Emissions')

    df = df[attributes]
    cl = ClusterClass(data=df, nb_clusters=[cluster['Periods']], option={"year-to-day": True, "extreme": []}, pd=cluster['PeriodDuration'])
    cl.run_clustering()

    generate_output_data(cl, attributes, cluster['Location'])


def get_weather_data(qbuildings_data):
    """
    Using the pvlib library, connects to the PVGIS dabatase to extract the weather data based on the building's coordinates.
    """
    lat, long = Transformer.from_crs("EPSG:2056", "EPSG:4326").transform(qbuildings_data['buildings_data']['Building1']['x'],
                                                                         qbuildings_data['buildings_data']['Building1']['y'])

    pvgis_data = pvlib.iotools.get_pvgis_tmy(lat, long, outputformat='csv', startyear=2005, endyear=2016)
    location = pvgis_data[2]
    weather_data = pvgis_data[0]
    weather_data = weather_data.rename(columns={'temp_air': 'Text', 'ghi': 'Irr'})
    weather_data['Month'] = weather_data.index.month
    weather_data['Day'] = weather_data.index.day
    weather_data['Hour'] = weather_data.index.hour + 1
    weather_data['id'] = (weather_data.reset_index().index + 1).to_list()
    weekday = pd.read_csv(os.path.join(path_to_weather, 'weekday.txt'), index_col=[0], header=None)
    weather_data['Weekday'] = weekday[1].tolist()

    print(f'The weather data have been loaded from the PVGIS database for a location with coordinates {location}.')

    return weather_data


def read_custom_weather(path_to_weather_file):
    """
    From the current directory, looks for a custom weather file.
    This file should be a *.csv* with the same structure as the template provided in reho/scripts/template/data/profiles/.
    """

    df = file_reader(path_handler(path_to_weather_file))
    print(f'The weather data have been loaded from {path_handler(path_to_weather_file)}.')

    df2 = pd.read_csv(os.path.join(path_to_weather, 'weekday.txt'), index_col=[0], header=None)
    df['Weekday'] = df2

    return df


def generate_output_data(cl, attributes, location):
    """
    Generates the data for the cluster timesteps obtained from the ClusterClass.

    Calls for *write_dat_files* to generate the saves.

    Parameters
    ----------
    cl : ClusterClass
        A ClusterClass object where the run_clustering method has already been executed.
    attributes : list
        Contains string among 'Irr', 'Text', 'Weekday'.
    location : str
        Location of the corresponding weather data.

    Notes
    ------
    .. caution::

        The extreme temperatures and irradiance are estimated by adding 10% to the extreme found in the yearly weather data.

    See also
    --------
    reho.model.preprocessing.clustering.ClusterClass.run_clustering
    write_dat_files
    """

    data_idx = cl.results["idx"]
    # - construct : cluster data
    frame = []
    cl.nbr_opt = str(cl.nbr_opt)
    for idx in data_idx.loc[:, cl.nbr_opt].unique():  # get unique typical periods from index vector
        idx = int(idx)
        df = pd.DataFrame(
            np.reshape(cl.attr_org[idx - 1, :], (-1, int(cl.attr_org.shape[1] / len(attributes)))).transpose(),
            columns=attributes)  # put the attributes into columns of a df
        df["dt"] = sum(data_idx.loc[:, cl.nbr_opt] == idx)  # Frequency
        df["time.hh"] = np.arange(1, df.shape[0] + 1, 1)  # timesteps in typical period
        df["time.dd"] = idx  # typical period index
        frame.append(df.reindex(["time.dd", "time.hh"] + attributes + ["dt"], axis=1))

    data_cls = pd.concat(frame, axis=0)
    data_cls_mod = pd.DataFrame()
    if cl.modulo != 0:  # Not clear what is this
        df_mod = pd.DataFrame.from_dict(dict(zip(cl.data_org.columns, cl.mod_org)))
        max_time_dd = len(cl.attr_org)
        df_mod['time.dd'] = np.repeat(max_time_dd + 1, cl.modulo)
        df_mod['dt'] = np.repeat(1, cl.modulo)
        df_mod['time.hh'] = np.arange(1, cl.modulo + 1)
        data_cls_mod = df_mod
    data_cls = pd.concat([data_cls, data_cls_mod], ignore_index=True)

    # Determine the day of the max and min temperature
    T_idx = [cl.data_org[cl.data_org['Text'] == cl.data_org['Text'].min()].index[0],
             cl.data_org[cl.data_org['Text'] == cl.data_org['Text'].max()].index[0]]
    T_day = [math.floor(T_idx[0] / 24), math.floor(T_idx[1] / 24)]
    T_min = cl.data_org.iloc[[T_idx[0]]].copy()
    # Get the max/min irradiance from the same day
    T_max = cl.data_org.iloc[[T_idx[1]]].copy()
    T_max['Irr'] = cl.data_org.loc[T_day[1] * 24: T_day[1] * 24 + 24, 'Irr'].max()
    T_min.loc[:, ['time.dd', 'time.hh', 'dt']] = [T_day[0], 1, 1]
    T_max.loc[:, ['time.dd', 'time.hh', 'dt']] = [T_day[1], 1, 1]
    data_cls = pd.concat([data_cls, T_min.rename({T_idx[0]: 240}), T_max.rename({T_idx[1]: 241})])
    # Add a 10% margin for the extreme over 20 years
    data_cls.loc[[240, 241], ['Text', 'Irr']] = data_cls.loc[[240, 241], ['Text', 'Irr']] * 1.1

    # - construct : model data
    # - ** inter-period
    data_idy = pd.DataFrame(
        np.stack((np.arange(1, data_idx.loc[:, cl.nbr_opt].shape[0] + 1, 1), data_idx.loc[:, cl.nbr_opt].values),
                 axis=1), columns=["IndexYr", "inter_t"])
    if cl.modulo != 0:
        max_time_dd = len(cl.attr_org)
        data_idy = pd.concat([data_idy, pd.DataFrame([[max_time_dd + 1, max_time_dd + 1]], columns=data_idy.columns)],
                             ignore_index=True)

    # - ** intra-period
    data_idp = pd.DataFrame(
        np.stack((np.arange(1, data_cls.shape[0] + 1, 1), np.arange(1, data_cls.shape[0] + 1, 1)), axis=1),
        columns=["IndexDy", "intra_t"])
    data_idp["intra_end"] = [id + cl.pd if (id % cl.pd) == 0 else 0 for id in data_idp.index]

    # Call for the write_dat_files function
    write_dat_files(attributes, location, data_cls, data_idy)

    return print(f'The data have been computed and saved in {path_to_clustering}.')


def write_dat_files(attributes, location, values_cluster, index_inter):
    """
    Writes the clustering results computed from `generate_output_data` as .dat file

    Parameters
    ----------
    attributes : list
        List that contains string among 'Irr', 'Text', 'Weekday'.
        If 'Irr' is in the list, writes a file named 'Irr' + '_File_ID.dat'
        If 'Text' is in the list, writes a file named 'T' + '_File_ID.dat'
    location : str
        Location of the corresponding weather data.
    values_cluster : pd.DataFrame
        Produced by the *generate_output_data* function.
    index_inter : pd.DataFrame
        Produced by the *generate_output_data* function.

    Notes
    -----
    - Not depending on the attributes, time dependent files are generated, namely 'frequency' + '_File_ID.dat',
      'index' + '_File_ID.dat' and 'timestamp' + '_File_ID.dat'.

    """

    df_dd = values_cluster['time.dd'].unique()  # id of typical period

    dp = np.array([])  # duration of period e.g. frequency
    pt = np.array([])  # period duration / number of timesteps in period

    for dd in df_dd:
        p = values_cluster.loc[values_cluster['time.dd'] == dd, 'dt'].unique()
        t = len(values_cluster.loc[values_cluster['time.dd'] == dd, 'dt'])

        dp = np.append(dp, p)
        pt = np.append(pt, t)

    # -------------------------------------------------------------------------------------
    # attributes for saving
    # -------------------------------------------------------------------------------------
    if 'Text' in attributes:
        T = '_T'
    else:
        T = ''
    if 'Irr' in attributes:
        I = '_I'
    else:
        I = ''
    if 'Weekday' in attributes:
        W = '_W'
    else:
        W = ''
    if 'Emissions' in attributes:
        E = '_E'
    else:
        E = ''
    File_ID = location + '_' + str(len(df_dd) - 2) + '_' + str(int(max(pt))) + T + I + W + E

    if not os.path.isdir(path_to_clustering):
        os.makedirs(path_to_clustering)

    # -------------------------------------------------------------------------------------
    # T
    # -------------------------------------------------------------------------------------
    df_T = values_cluster['Text']
    filename = os.path.join(path_to_clustering, 'T_' + File_ID + '.dat')
    df_T.to_csv(filename, index=False, header=False)

    # -------------------------------------------------------------------------------------
    # Irr
    # -------------------------------------------------------------------------------------
    df_Irr = values_cluster['Irr']
    filename = os.path.join(path_to_clustering, 'Irr_' + File_ID + '.dat')
    df_Irr.to_csv(filename, index=False, header=False)

    # -------------------------------------------------------------------------------------
    # frequency
    # -------------------------------------------------------------------------------------
    if 'Weekday' in attributes:
        Weekday = np.array([])
        for dd in df_dd:
            w = values_cluster.loc[values_cluster['time.dd'] == dd, 'Weekday'].unique()
            Weekday = np.append(Weekday, w)

    filename = os.path.join(path_to_clustering, 'frequency_' + File_ID + '.dat')

    IterationFile = open(filename, 'w')

    IterationFile.write('\nset Period := ')
    for p in range(1, len(dp) + 1):  # +1 bc ampl starts at 0, +2 for extreme periods
        IterationFile.write('\n' + str(p))
    IterationFile.write('\n;')

    IterationFile.write('\nset PeriodStandard := ')
    for p in range(1, len(dp) - 1):  # +1 bc ampl starts at 0, -2 to exclude extreme periods
        IterationFile.write('\n' + str(p))
    IterationFile.write('\n;')

    IterationFile.write('\nparam: dp := ')
    for p, d in enumerate(dp):
        IterationFile.write('\n' + str(p + 1) + ' ' + str(d))
    IterationFile.write('\n;')

    IterationFile.write('\nparam: TimeEnd := ')
    for p, d in enumerate(pt):
        IterationFile.write('\n' + str(p + 1) + ' ' + str(d))
    IterationFile.write('\n;')
    IterationFile.close()

    # -------------------------------------------------------------------------------------
    # index
    # -------------------------------------------------------------------------------------
    df_time = pd.DataFrame()
    df_time['originalday'] = df_dd
    df_time['frequency'] = dp
    df_time['timesteps'] = pt
    dict_index = {}
    for i, dd in enumerate(df_dd):
        dict_index[dd] = i + 1
    index_inter['index_r'] = index_inter.inter_t.map(dict_index)
    df_time.index = df_time.index + 1

    df_aim = pd.DataFrame()
    for d in index_inter['index_r']:
        nt = int(df_time['timesteps'].xs(d))  # number of timesteps
        df_d = pd.DataFrame([np.repeat(d, nt), np.array(range(1, (nt + 1)))])
        df_aim = pd.concat([df_aim, df_d.transpose()], ignore_index=True)

    df_aim.index = df_aim.index + 1

    filename = os.path.join(path_to_clustering, 'index_' + File_ID + '.dat')
    IterationFile = open(filename, 'w')
    IterationFile.write('param : PeriodOfYear TimeOfYear := \n')
    IterationFile.write(df_aim.to_string(header=False))
    IterationFile.write('\n;')
    IterationFile.close()

    # -------------------------------------------------------------------------------------
    # Time stamp
    # -------------------------------------------------------------------------------------
    filename = os.path.join(path_to_clustering, 'timestamp_' + File_ID + '.dat')
    IterationFile = open(filename, 'w')
    header = 'Date\tDay\tFrequency\tWeekday\n'
    IterationFile.write(header)
    for key in dict_index:
        pt = df_time.iloc[0].timesteps  # take the same period duration also for modulo
        date = dt.datetime(2005, 1, 1) + dt.timedelta(hours=float((key - 1) * pt))

        if 'Weekday' in attributes:
            text = date.strftime("%m/%d/%Y/%H") + '\t' + str(key) + '\t' + str(dp[dict_index[key] - 1]) + '\t' + str(
                Weekday[dict_index[key] - 1])
        else:
            text = date.strftime("%m/%d/%Y/%H") + '\t' + str(key) + '\t' + str(dp[dict_index[key] - 1])
        IterationFile.write(text + '\n')
    IterationFile.close()

    return


def plot_cluster_KPI_separate(df, save_fig):
    plt.rcParams.update({'font.size': 18})
    print(df)
    df = df.transpose()
    df.columns.names = ['KPI']
    df.rename({'Irr': 'Global Irradiation', 'Text': 'Ambient Temperature'}, inplace=True)
    df_irr = df.xs('Global Irradiation', level=1)
    df_T = df.xs('Ambient Temperature', level=1)

    fig, ax = plt.subplots()
    fig.set_size_inches(4, 8)
    df_irr['RMSD'].plot(linestyle='--', color='black', label='RMSD (Irr)', ax=ax)
    df_T['RMSD'].plot(linestyle='-', color='black', label='RMSD (T)', ax=ax)
    df_irr['LDC'].plot(linestyle='--', color="red", label='LDC (Irr)', ax=ax)
    df_T['LDC'].plot(linestyle='-', color="red", label='LDC (T)', ax=ax)

    plt.xlabel('number of clusters [-]')
    plt.ylabel('key performance indicator (KPI) [-]')
    # plt.title('KPI for $ \u2B27 $ =  Global Irradiation, $\u00D7$ = Ambient Temperature', size = 14)
    plt.legend(title="KPI")

    # plt.ylim([0,0.40])
    if save_fig:
        plt.tight_layout()
        export_format = 'pdf'
        plt.savefig(('Cluster_KPIs' + '.' + export_format), format=export_format, dpi=300)
    else:
        plt.show()

    fig, ax = plt.subplots()
    fig.set_size_inches(4, 8)
    df_irr['MAE'].plot(linestyle='--', color='black', label='MAE (Irr)', ax=ax)
    df_T['MAE'].plot(linestyle='-', color='black', label='MAE (T)', ax=ax)

    plt.xlabel('number of clusters [-]')
    plt.ylabel('mean average error (MAE)  [-]')
    # plt.title('KPI for $ \u2B27 $ =  Global Irradiation, $\u00D7$ = Ambient Temperature', size = 14)
    plt.legend(title="KPI")
    # plt.ylim([0,0.40])
    if save_fig:
        plt.tight_layout()
        export_format = 'pdf'
        plt.savefig(('MAE_KPIs' + '.' + export_format), format=export_format, dpi=300)
    else:
        plt.show()

    fig, ax = plt.subplots()
    fig.set_size_inches(4, 8)
    df_irr['MAPE'].plot(linestyle='--', color='black', label='MAPE  (Irr)', ax=ax)
    df_T['MAPE'].plot(linestyle='-', color='black', label='MAPE  (T)', ax=ax)

    plt.xlabel('number of clusters [-]')
    plt.ylabel('mean average percentage error  [-]')
    # plt.title('KPI for $ \u2B27 $ =  Global Irradiation, $\u00D7$ = Ambient Temperature', size = 14)
    plt.legend(title="KPI")
    # plt.ylim([0,0.40])
    if save_fig:
        plt.tight_layout()
        export_format = 'pdf'
        plt.savefig(('MAPE_KPIs' + '.' + export_format), format=export_format, dpi=300)
    else:
        plt.show()


def plot_LDC(cl, save_fig):
    nbr_plot = cl.nbr_opt
    print('plotting for number of typical days: ', nbr_plot)

    # get original, not clustered data
    T_org = cl.data_org['Text']
    IRR_org = cl.data_org['Irr']
    # E_org= cl.data_org['Emissions']

    # get clustered data and undo normalization
    df_clu = cl.attr_clu.xs(str(nbr_plot), axis=1)
    T_clu = df_clu['Text'] * (T_org.max() - T_org.min()) + T_org.min()
    IRR_clu = df_clu['Irr'] * (IRR_org.max() - IRR_org.min()) + IRR_org.min()
    #  E_clu= df_clu['Emissions']* (E_org.max() - E_org.min()) + E_org.min()

    # get assigned typical period
    res = cl.results['idx'][str(nbr_plot)]
    # instead of typical period (f.e. day 340) get index (1)
    for i, d in enumerate(res.unique()):
        res = np.where(res == d, i + 1, res)
    # get array of 8760 -modulo timesteps
    res = np.repeat(res, cl.pd)

    # add modulo
    modulo = cl.data_org.shape[0] % cl.pd
    res = np.append(res, np.repeat(int(nbr_plot) + 1, modulo))  # add modulo as extra period

    # ----------------------------------------------------------------------------
    # Plotting
    # ------------------------------------------------------------------------

    fig, ax = plt.subplots(2, 1, sharex=True, figsize=(10, 8))

    ax[0].plot(T_org, color='grey', alpha=0.5)
    sc = ax[0].scatter(T_clu.index, T_clu.values, s=10, c=res, cmap=cm)
    ax[1].plot(IRR_org, color='grey', alpha=0.5)
    ax[1].scatter(IRR_clu.index, IRR_clu.values, s=10, c=res, cmap=cm)
    # ax[2].plot(E_org, color='grey', alpha=0.5)
    # ax[2].scatter(E_clu.index, E_clu.values, s=10, c=res, cmap=cm)

    # set months instead of timestep as xticks
    plt.xticks(np.arange(8760, step=730), calendar.month_name[1:13], rotation=30, fontsize=16)

    ax[0].set_ylabel('temperature [C]')
    ax[1].set_ylabel('global irradiation [W/m$^2$]')
    # ax[2].set_ylabel('global warming potential [gCO2/kWh]')
    # plt.subplots_adjust(bottom=0.1, right=0.8, top=0.9)

    legend1 = ax[0].legend(*sc.legend_elements(),
                           loc='upper right', bbox_to_anchor=(1.0, 1.0), title="Period", ncol=2)
    ax[0].add_artist(legend1)

    if save_fig:
        plt.tight_layout()
        format = 'pdf'
        plt.savefig(('Year_Cluster' + '.' + format), format=format, dpi=300)
    else:
        plt.show()

    df_T = pd.DataFrame(T_clu)
    df_T['Period'] = res
    df_T = df_T.sort_values(by=['Text'], ignore_index=True, ascending=False)

    df_Irr = pd.DataFrame(IRR_clu)
    df_Irr['Period'] = res
    df_Irr = df_Irr.sort_values(by=['Irr'], ignore_index=True, ascending=False)

    # df_E = pd.DataFrame(E_clu)
    # df_E['Period'] = res
    # df_E = df_E.sort_values(by=['Emissions'], ignore_index=True, ascending=False)

    T_sort = T_org.sort_values(ascending=False, ignore_index=True)
    IRR_sort = IRR_org.sort_values(ascending=False, ignore_index=True)
    # E_sort =  E_org.sort_values(ascending=False, ignore_index=True)

    fig, ax = plt.subplots(2, 1, sharex=True, figsize=(10, 8))
    ax[0].scatter(T_sort.index, T_sort.values, color='grey', alpha=0.5)
    ax[0].scatter(df_T.index, df_T['Text'], c=df_T['Period'], cmap=cm, s=20)

    ax[1].scatter(IRR_sort.index, IRR_sort.values, color='grey', alpha=0.5)
    ax[1].scatter(df_Irr.index, df_Irr['Irr'], c=df_Irr['Period'], cmap=cm, s=20)

    #  ax[2].scatter(E_sort.index, E_sort.values, color='grey', alpha=0.5)
    # ax[2].scatter(df_E.index, df_E['Emissions'], c=df_Irr['Period'], cmap=cm, s=20)

    ax[0].set_ylabel('temperature [C]')
    ax[1].set_ylabel('global irradiation [W/m$^2$]')
    # ax[2].set_ylabel('global warming potential [gCO2/kWh]')

    plt.xlabel('Hours [h]')

    legend1 = ax[0].legend(*sc.legend_elements(),
                           loc='upper right', bbox_to_anchor=(1.0, 1.0), title="Period", ncol=2)
    ax[0].add_artist(legend1)

    if save_fig:
        plt.tight_layout()
        format = 'pdf'
        plt.savefig(('LDC' + '.' + format), format=format, dpi=300)
    else:
        plt.show()


if __name__ == '__main__':
    cm = plt.cm.get_cmap('Spectral_r')

    weather_file = '../../../scripts/template/data/profiles/Sion.csv'
    Attributes = ['Text', 'Irr']
    nb_clusters = [2, 4, 6, 8, 10, 12]

    df_annual = read_custom_weather(weather_file)
    df_annual = df_annual[Attributes]

    cl = ClusterClass(data=df_annual, nb_clusters=nb_clusters, option={"year-to-day": True, "extreme": []}, pd=24)
    cl.run_clustering()

    plot_cluster_KPI_separate(cl.kpis_clu, save_fig=False)
    plot_LDC(cl, save_fig=False)

    generate_output_data(cl, Attributes, "Sion")



import matplotlib.pyplot as plt
from SALib.analyze import morris as morris_analyze
from SALib.analyze import sobol as sobol_analyze
from SALib.sample import morris as morris_sample
from SALib.sample import sobol as sobol_sample
from qmcpy import Sobol

from reho.model.reho import *


class SensitivityAnalysis:
    """
    Performs a sensitivity analysis (SA): sampling, problem, store all optimizations results and the
    sensitivity of each tested parameters.

    Parameters
    ----------
    reho : reho object
        Model of the district, obtained via the REHO class.
    SA_type : str
        Type of SA (Morris or Sobol).
    sampling_parameters : int
        Number of trajectories for the sampling of the solution space.
    upscaling_factor : int
        To represent the effective ERA of the typical districts.

    Notes
    -------
    The framework is designed to be performed using TOTEX minimization but can easily be modified,
    just change the objective function of the reho and the KPI saved into OBJ in the function run_SA()
    """

    def __init__(self, reho, SA_type, sampling_parameters=0, upscaling_factor=1):

        self.district_name = reho.scenario["name"]
        self.reho = reho
        self.SA_type = SA_type
        self.upscaling_factor = upscaling_factor
        self.sampling_parameters = sampling_parameters
        self.ID = str(self.district_name) + "_" + str(self.SA_type) + time.strftime("_%d%m%Y@%Hh%M")
        self.parameter = {}
        self.problem = {}
        self.sampling = []
        self.OBJ = []
        self.SA_results = {'num_optimizations': [], 'dict_df_results': [], 'dict_res_ES': []}
        self.sensitivity = []

    def save(self):
        path_to_SA_results = 'results/'
        file_path = os.path.join(path_to_SA_results, self.ID + '.pickle')
        f = open(file_path, 'wb')
        pickle.dump(self, f)
        f.close()

    def get_lists(self):
        unit_list = self.reho.infrastructure.UnitTypes.tolist()
        try:
            for excluded_unit in self.reho.scenario['exclude_units']:
                unit_list.remove(excluded_unit)
        except ValueError:
            pass
        KPI_list = ['OPEX', 'CAPEX', 'TOTEX', 'GWP']
        return unit_list, KPI_list

    def build_SA(self, unit_parameter=['Cost_inv1', 'Cost_inv2'], SA_parameters={}):
        """
        Description:
        - Generate the list of parameters for the SA, their values and type of variation range
        - Generate the problem of the SA, i.e. define the parameters and theirs bounds
        - Generate the sampling scheme of the SA

        Parameters
        ----------
        unit_parameter [list]: Units parameters wanted in the SA, by default all parameters are included
        SA_parameters [dict]: Parameters with their bounds set by the script

        Returns
        ---------
        parameter : dict
            Parameters
        problem : dict
            Parameters with their bounds
        sampling : array
            Sampling values
        """

        # 1) Generate the list of parameters

        default_units_values = self.reho.infrastructure.Units_Parameters  # Extract default unit values of the district
        units = np.unique([unit.split('_Building')[0] for unit in self.reho.infrastructure.Units_Parameters.index.to_list()]).tolist()
        if "EV_district" in units:
            units.remove("EV_district")

        for item in self.reho.scenario['exclude_units']:  # Removing excluded units
            try:
                units.remove(item)
            except ValueError:
                pass

        if unit_parameter != []:  # Add units parameters
            for unit in units:
                for parameter in unit_parameter:
                    value = default_units_values[default_units_values.index.str.contains(unit)][parameter].iloc[0]
                    name = str(unit) + "___" + str(parameter)
                    SA_parameters[name] = np.array([0.5, 2]) * value

        self.parameter = SA_parameters
        # 2) Generate a dictionary with all parameters and their bounds for the sampling

        problem = {
            'names': list(SA_parameters.keys()),
            'num_vars': len(SA_parameters.keys()),
            'bounds': list(SA_parameters.values())
        }
        self.problem = problem

        # 3) Description: Generate the sampling of the solution space

        if self.SA_type == "Sobol":
            sampling = sobol_sample.sample(self.problem, self.sampling_parameters, calc_second_order=False)  # Between 500 and 1000, ideally a power of 2
        elif self.SA_type == "Morris":
            sampling = morris_sample.sample(self.problem, self.sampling_parameters)  # Sampling of the space, recommended p = 4 (level) r = 10 (trajectories)
        elif self.SA_type == "Monte_Carlo":
            dimension = len(self.problem["bounds"])
            sampler = Sobol(dimension)
            sample = sampler.gen_samples(self.sampling_parameters)
            l_bounds_values = [bound[0] for bound in self.problem["bounds"]]
            u_bounds_values = [bound[1] for bound in self.problem["bounds"]]
            sampling = qmc.scale(sample, l_bounds_values, u_bounds_values)
        else:
            sampling = None
        self.sampling = sampling

    def run_SA(self, save_inter=True, save_inter_nb_iter=50, save_time_opt=True, intermediate_start=0):
        """
        Launches all optimizations of the SA and store their results.

        Parameters
        -----------

        save_inter : boolean
            Enable intermediary save
        save_inter_nb_iter : int
            Step at which the intermediary save is done
        save_time_opt : boolean
            Ceates a .txt file and write the time for each optimization
        intermediate_start :int
            Starts the SA from a specific sampling point

        Returns
        ---------
        SA_results : dict
            Contains the number of the optimization and a dictionary regrouping all main results of the optimizations
        OBJ : list
            Values of the objective function for each optimization
        """

        path_to_SA_results = 'results/'
        folder = os.path.join(path_to_SA_results, "computational_results")
        if not os.path.exists(path_to_SA_results):
            os.makedirs(path_to_SA_results)
        if not os.path.exists(folder):
            os.makedirs(folder)

        # Extract all attributes of the reho model
        grids = self.reho.infrastructure.grids
        scenario = self.reho.scenario
        district_units = len(self.reho.infrastructure.UnitsOfDistrict) != 0  # True or False
        units = infrastructure.initialize_units(scenario, grids, district_data=district_units)
        n_houses = len(self.reho.buildings_data)

        # Modify the attributes of the model and run SA
        for j in range(intermediate_start, len(self.sampling)):
            print("Optimization number", str(j + 1) + "/" + str(len(self.sampling)))

            sample = self.sampling[j]
            for s, value in enumerate(sample):
                parameter = list(self.parameter.keys())[s]

                if parameter == 'Elec_retail':
                    grids["Electricity"]["Cost_supply_cst"] = value
                elif parameter == 'Elec_feedin':
                    grids["Electricity"]["Cost_demand_cst"] = value
                elif parameter == 'NG_retail':
                    grids["NaturalGas"]["Cost_supply_cst"] = value
                elif parameter == 'Wood_retail':
                    grids["Wood"]["Cost_supply_cst"] = value
                elif parameter == 'Oil_retail':
                    grids["Oil"]["Cost_supply_cst"] = value

                elif "___" in parameter:
                    for unit_id in range(len(units['building_units'])):
                        if units['building_units'][unit_id]['name'] == parameter.split("___")[0]:
                            units['building_units'][unit_id][parameter.split("___")[1]] = value
                else:
                    if parameter in self.reho.lists_MP["list_parameters_MP"]:
                        self.reho.parameters[parameter] = np.array([value])
                    else:
                        self.reho.parameters[parameter] = np.array([value] * n_houses)

            qbuildings_data = {'buildings_data': self.reho.buildings_data}
            self.reho.infrastructure = infrastructure.Infrastructure(qbuildings_data, units, grids)

            try:
                tic = time.perf_counter()
                self.reho.single_optimization(Pareto_ID=j)  # Optimize the modified model
                toc = time.perf_counter()
                time_spent = toc - tic

                self.reho.initialize_optimization_tracking_attributes()

                if save_time_opt:
                    file_name = os.path.join(folder, str(self.SA_type) + '_time.txt')
                    function = 'w' if not os.path.exists(folder) else 'a+'
                    with open(file_name, function) as f:
                        f.write(str(round(time_spent)) + "\n")

                if save_inter:  # Intermediary save
                    if np.mod(j, save_inter_nb_iter) == 0 or j == (len(self.sampling) - 1):
                        self.save()
                        os.system('cmd /c "ampl_lic restart"')  # restart ampl license to avoid crashes

            except KeyboardInterrupt:
                return

    def calculate_SA(self):
        """
        Computes the sensitivity indices with the objective values and the problem.
        """
        if self.SA_type == "Sobol":
            sensitivity = sobol_analyze.analyze(self.problem, np.array(self.OBJ), print_to_console=True, calc_second_order=False)
        if self.SA_type == "Morris":
            sensitivity = morris_analyze.analyze(self.problem, self.sampling, np.array(self.OBJ), print_to_console=True)
        self.sensitivity = sensitivity

    def plot_Morris(self, save=False):
        fig, ax = plt.subplots(figsize=(8, 8))
        df_ = pd.DataFrame(self.sensitivity).sort_values('mu_star', ascending=False).copy()
        df_.reset_index(inplace=True, drop=True)
        for j, row in df_.iterrows():
            plt.scatter(row['mu_star'], y=row['sigma'], label=row["names"], c='blue', alpha=0.3)
            if row.name < 5:
                plt.annotate(row["names"], xy=(row["mu_star"], row["sigma"]), textcoords="offset pixels",
                             xytext=(20, 20), arrowprops=dict(arrowstyle="-", connectionstyle="arc3"), fontsize=15)
        plt.plot(ax.get_xlim(), ax.get_xlim(), linestyle='dotted', color='black')
        plt.xlabel('$\\mu^*$ [-]', fontsize=18)
        plt.ylabel('$\\sigma$ [-]', fontsize=18)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.title(self.district_name, fontsize=25)
        plt.axis('square')

        if save:
            path_to_SA_results = 'results/'
            plt.savefig(path_to_SA_results + 'Morris_' + self.district_name + '.png', format='png', dpi=300)
        plt.show()

    def extract_results(self, reho, j):
        unit_list, KPI_list = self.get_lists()

        dict_res = {
            'Annual_Network_Exchange': reho.results[self.SA_type][0].df_Annuals.xs("Network", level=1).loc[["Electricity", "NaturalGas"]][
                ['Demand_MWh', 'Supply_MWh']],
            'Elec_Network_t': reho.results[self.SA_type][0].df_Grid_t.xs("Network", level="Hub").xs("Electricity")[['Grid_demand', 'Grid_supply']],
            'NG_Network_t': reho.results[self.SA_type][0].df_Grid_t.xs("Network", level="Hub").xs("NaturalGas")[['Grid_demand', 'Grid_supply']],
            'df_Unit': reho.results[self.SA_type][0].df_Unit['Units_Mult'],
            'Performance_Network': reho.results[self.SA_type][0].df_Performance.xs("Network")
        }

        self.SA_results['num_optimizations'].append(j)
        self.SA_results['dict_df_results'].append(dict_res)
        self.OBJ.append(reho.results[self.SA_type][0].df_Performance['Costs_inv']['Network']
                        + reho.results[self.SA_type][0].df_Performance['Costs_op']['Network']
                        + reho.results[self.SA_type][0].df_Performance['Costs_rep']['Network'])

        df_Grid_t = reho.results[self.SA_type][0].df_Grid_t[['Grid_demand', 'Grid_supply']].groupby(['Layer', 'Hub', 'Period']).sum()
        df_Annuals = reho.results[self.SA_type][0].df_Annuals
        df_Unit_t = reho.results[self.SA_type][0].df_Unit_t.groupby(['Layer', 'Unit', 'Period']).sum()

        dict_res_ES = {'E_sector': df_Annuals.groupby(['Layer']).sum(),
                       'InAndOutDistrict': df_Grid_t.xs("Network", level="Hub"),
                       'E_unit': {},
                       'E_resource': df_Unit_t[['Units_demand', 'Units_supply']].groupby(['Layer', 'Period']).sum(),
                       'E_unit_PV': df_Unit_t.xs("Electricity", level="Layer").query('Unit.str.startswith("PV")').groupby(['Period']).sum()['Units_supply']}

        for unit in unit_list:
            dict_res_ES['E_unit'][unit] = df_Annuals.query('Hub.str.startswith("' + str(unit) + '")')['Supply_MWh'].values.sum()

        self.SA_results['dict_res_ES'].append(dict_res_ES)



import pandas as pd
import numpy as np
import logging

__doc__ = """
Extracts the results from the AMPL model and converts it to Python dictionary and pandas dataframes.
"""


def get_df_Results_from_SP(ampl, scenario, method, buildings_data, filter=True):
    def set_df_performance(ampl, scenario):
        df1 = get_ampl_data(ampl, 'Costs_House_op')  # without the comfort penalty costs
        df1 = df1.rename(columns={'Costs_House_op': 'Costs_op'})

        df2 = get_ampl_data(ampl, 'Costs_House_inv')
        df2 = df2.rename(columns={'Costs_House_inv': 'Costs_inv'})
        tau = ampl.getParameter('tau').getValues().toList()
        df2['Costs_inv'] = df2['Costs_inv'] * tau[0]
        df2['ANN_factor'] = tau[0]

        df2['Costs_grid_connection'] = get_ampl_data(ampl, 'Costs_grid_connection_House', multi_index=True).groupby(
            level=1).sum()  # yearly cost for grid connection

        df3 = get_ampl_data(ampl, 'Costs_House_rep')
        df3['Costs_House_rep'] = df3['Costs_House_rep'] * tau[0]
        df3 = df3.rename(columns={'Costs_House_rep': 'Costs_rep'})

        df4 = get_ampl_data(ampl, 'Costs_House_cft')
        df4 = df4.rename(columns={'Costs_House_cft': 'Costs_ft'})

        df5 = get_ampl_data(ampl, 'GWP_house_op')
        df5 = df5.rename(columns={'GWP_house_op': 'GWP_op'})

        df6 = get_ampl_data(ampl, 'GWP_house_constr')
        df6 = df6.rename(columns={'GWP_house_constr': 'GWP_constr'})

        df71 = get_ampl_data(ampl, 'EMOO_CAPEX')
        df72 = get_ampl_data(ampl, 'EMOO_OPEX')
        df73 = get_ampl_data(ampl, 'EMOO_TOTEX')
        df75 = get_ampl_data(ampl, 'EMOO_GWP')
        df76 = get_ampl_data(ampl, 'EMOO_grid')

        df_N1 = get_ampl_data(ampl, 'Costs_op')  # without the comfort penalty costs
        df_N2 = get_ampl_data(ampl, 'Costs_inv')
        df_N2['Costs_inv'] = df_N2['Costs_inv'] * tau[0]
        df_N2['ANN_factor'] = tau[0]
        df_N2['Costs_grid_connection'] = get_ampl_data(ampl, 'Costs_grid_connection').sum().values / 2  # TODO enhance
        df_N3 = get_ampl_data(ampl, 'Costs_rep')
        df_N3['Costs_rep'] = df_N3['Costs_rep'] * tau[0]
        df_N4 = pd.DataFrame({'Costs_ft': [df4.sum()['Costs_ft']]})
        df_N5 = get_ampl_data(ampl, 'GWP_op')
        df_N6 = get_ampl_data(ampl, 'GWP_constr')

        df_PerformanceBuilding = pd.concat([df1, df2, df3, df4, df5, df6], axis=1)
        df_PerformanceNetwork = pd.concat([df_N1, df_N2, df_N3, df_N4, df_N5, df_N6], axis=1)
        df_PerformanceNetwork = df_PerformanceNetwork.rename(index={0: 'Network'})

        df_Performance = pd.concat([df_PerformanceBuilding, df_PerformanceNetwork], axis=0)

        df_Epsilon = pd.concat([df71, df72, df73, df75, df76], axis=1)
        df_Epsilon['Objective'] = get_ampl_data(ampl, scenario["Objective"]).values[0][0] - df_N4.values[0][0]
        df_Epsilon = df_Epsilon.rename(index={0: 'Network'})

        df_Performance = pd.concat([df_Performance, df_Epsilon], axis=1)
        df_Performance.index.names = ['Hub']
        if method['print_logs']:
            print(df_Performance)

        return df_Performance.sort_index()

    def set_df_annuals(ampl):
        # Annuals
        df1 = get_ampl_data(ampl, 'AnnualNetwork_demand')
        df1.columns = ['Demand_MWh']
        df1 = pd.concat([df1], keys=['Network'])
        df1.index = df1.index.reorder_levels([1, 0])
        df2 = get_ampl_data(ampl, 'AnnualNetwork_supply')
        df2.columns = ['Supply_MWh']
        df2 = pd.concat([df2], keys=['Network'])
        df2.index = df2.index.reorder_levels([1, 0])
        df12 = pd.concat([df1, df2], axis=1, sort=False)

        df3 = get_ampl_data(ampl, 'AnnualDomestic_electricity')
        df3 = df3.set_index([pd.Index(["Electricity"] * df3.index.size), df3.index])
        df3.columns = ['Demand_MWh']
        df4 = get_ampl_data(ampl, 'AnnualHouse_Q')
        df4.columns = ['Demand_MWh']
        df5 = get_ampl_data(ampl, 'AnnualHeatGainHouse')
        df5.columns = ['Supply_MWh']
        df5 = pd.concat([df5], keys=['HeatGains'])
        df6 = get_ampl_data(ampl, 'AnnualSolarGainHouse')
        df6.columns = ['Supply_MWh']
        df6 = pd.concat([df6], keys=['SolarGains'])
        df3456 = pd.concat([df3, df4, df5, df6], sort=False)

        df7 = get_ampl_data(ampl, 'AnnualUnit_in')
        df7.columns = ['Demand_MWh']
        df8 = get_ampl_data(ampl, 'AnnualUnit_out')
        df8.columns = ['Supply_MWh']
        df78 = pd.concat([df7, df8], axis=1, sort=False)
        df9 = get_ampl_data(ampl, 'AnnualUnit_Q')
        df9.columns = ['Supply_MWh']
        df789 = pd.concat([df78, df9], sort=False)

        df_Annuals = pd.concat([df12, df3456, df789], sort=False)
        df_Annuals.index.names = ['Layer', 'Hub']

        # Correction of DHW layer
        hubs = [s for s in df_Annuals.index.levels[1] if s.startswith("Building")]
        for h in hubs:
            df_Annuals.loc[('DHW', h), 'Demand_MWh'] = df_Annuals.loc[('DHW', 'WaterTankDHW_' + h), 'Supply_MWh']

        return df_Annuals

    def set_df_buildings(buildings_data):
        # Building
        df_Buildings = pd.DataFrame.from_dict(buildings_data, orient='index')
        df_Buildings.index.names = ['Hub']
        for item in ['x', 'y', 'z', 'geometry']:
            if item in df_Buildings.columns:
                df_Buildings.drop([item], axis=1)

        return df_Buildings.sort_index()

    def set_df_unit(ampl):
        # Unit
        tau = ampl.getParameter('tau').getValues().toList()
        df1 = get_ampl_data(ampl, 'Units_Use')
        df2 = get_ampl_data(ampl, 'Units_Mult')

        df3 = tau[0] * get_ampl_data(ampl, 'Costs_Unit_inv')
        df4 = get_ampl_data(ampl, 'GWP_Unit_constr')  # per year! For total, multiply with lifetime
        df5 = get_ampl_data(ampl, 'lifetime')
        df_Unit = pd.concat([df1, df2, df3, df4, df5], axis=1)
        df_Unit.index.names = ['Unit']
        df_Unit = df_Unit.sort_index()
        if method['print_logs']:
            print(df_Unit)

        # Unit_t
        df1 = get_ampl_data(ampl, 'Units_demand', multi_index=True)
        df2 = get_ampl_data(ampl, 'Units_supply', multi_index=True)
        df3 = get_ampl_data(ampl, 'Units_curtailment', multi_index=True)
        df4 = get_ampl_data(ampl, 'BAT_E_stored', multi_index=True)
        df4 = pd.concat([df4], keys=['Electricity'], names=['Layer'])
        if "EV_district" in [unit for unit, value in ampl.getVariable('Units_Use').instances()]:
            df5 = get_ampl_data(ampl, 'EV_E_stored', multi_index=True)
            df5 = pd.concat([df5], keys=['Electricity'], names=['Layer'])
            df6 = get_ampl_data(ampl, "EV_displacement", multi_index=True)
            df6 = pd.concat([df6], keys=['Electricity'], names=['Layer'])
            df_Unit_t = pd.concat([df1, df2, df3, df4, df5, df6], axis=1)
        else:
            df_Unit_t = pd.concat([df1, df2, df3, df4], axis=1)
        df_Unit_t.index.names = ['Layer', 'Unit', 'Period', 'Time']
        df_Unit_t = df_Unit_t.sort_index()

        return df_Unit, df_Unit_t

    def set_df_grid(ampl, method):
        # Grid_t
        df1 = get_ampl_data(ampl, 'Grid_demand', multi_index=True)
        df2 = get_ampl_data(ampl, 'Grid_supply', multi_index=True)

        df_cs = get_ampl_data(ampl, 'Cost_supply', multi_index=True)
        df_cs = df_cs.reorder_levels((1, 0, 2, 3))
        df_cd = get_ampl_data(ampl, 'Cost_demand', multi_index=True)
        df_cd = df_cd.reorder_levels((1, 0, 2, 3))

        df_electricity = get_ampl_data(ampl, 'Domestic_electricity', multi_index=True)
        df_electricity.columns = ['Uncontrollable_load']
        df_electricity = df_electricity.set_index([pd.Index(["Electricity"] * df_electricity.index.size), df_electricity.index])

        df_es = get_ampl_data(ampl, 'GWP_supply', multi_index=True)
        df_ed = get_ampl_data(ampl, 'GWP_demand', multi_index=True)
        df_em = pd.concat([df_es, df_ed], axis=1)
        df_em_tot = pd.DataFrame()
        for bui in df1.index.get_level_values(1).unique():
            df = df_em
            df["Hub"] = bui
            df_em_tot = pd.concat([df_em_tot, df])
        df_em_tot.set_index('Hub', append=True, inplace=True)
        df_em_tot = df_em_tot.reorder_levels((0, 3, 1, 2))

        df_Grid_t = pd.concat([df1, df2, df_cs, df_cd, df_em_tot, df_electricity], axis=1)

        if not method["district-scale"] or not method["district-scale"]:
            df3 = get_ampl_data(ampl, 'Network_demand', multi_index=True)
            df3.columns = ['Grid_demand']
            df4 = get_ampl_data(ampl, 'Network_supply', multi_index=True)
            df4.columns = ['Grid_supply']
            df_cs_net = get_ampl_data(ampl, 'Cost_supply_network', multi_index=True)
            df_cs_net.columns = ['Cost_supply']
            df_cd_net = get_ampl_data(ampl, 'Cost_demand_network', multi_index=True)
            df_cd_net.columns = ['Cost_demand']

            df_electricity_net = df_electricity.groupby(level=(0, 2, 3)).sum()
            df34 = pd.concat([df3, df4, df_cs_net, df_cd_net, df_es, df_ed, df_electricity_net], axis=1)
            df34['Hub'] = 'Network'
            df34.set_index('Hub', append=True, inplace=True)
            df34 = df34.reorder_levels((0, 3, 1, 2))
            df_Grid_t = pd.concat([df_Grid_t, df34])

        df_Grid_t.index.names = ['Layer', 'Hub', 'Period', 'Time']
        return df_Grid_t.sort_index()

    def set_df_buildings_t(ampl):
        # Building_t
        df1 = get_ampl_data(ampl, 'Domestic_electricity', multi_index=True)

        m_DWH = get_ampl_data(ampl, 'DHW_flowrate', multi_index=True)
        delta_T = get_ampl_data(ampl, 'DHW_dT').iloc[0, 0]
        df2 = 4.18 * m_DWH * delta_T / 3600
        df2 = df2.rename(columns={'DHW_flowrate': 'House_Q_DHW'})

        df31 = get_ampl_data(ampl, 'T_in', multi_index=True)
        df32 = get_ampl_data(ampl, 'House_Q_heating')
        df33 = get_ampl_data(ampl, 'House_Q_cooling')
        df34 = get_ampl_data(ampl, 'Th_supply', multi_index=True)
        df35 = get_ampl_data(ampl, 'Th_return', multi_index=True)

        df4 = get_ampl_data(ampl, 'HeatGains', multi_index=True)
        df5 = get_ampl_data(ampl, 'SolarGains', multi_index=True)

        df_Buildings_t = pd.concat([df1, df2, df31, df32, df33, df34, df35, df4, df5], axis=1)
        df_Buildings_t.index.names = ['Hub', 'Period', 'Time']

        return df_Buildings_t.sort_index()

    def set_df_streams_t(ampl):

        df_Q = get_ampl_data(ampl, 'Streams_Q', multi_index=True)
        df_Q.index.set_names(['Service', 'Stream', 'Period', 'Time'], inplace=True)

        df1 = get_ampl_data(ampl, 'HC_Streams_Mult', multi_index=True)
        df1.index.set_names(['Service', 'Stream', 'Period', 'Time'], inplace=True)
        df2 = get_ampl_data(ampl, 'Streams_Mcp', multi_index=True)
        df2.index.set_names(['Stream', 'Period', 'Time'], inplace=True)
        df_mcp = df1['HC_Streams_Mult'].mul(df2['Streams_Mcp'])
        df_mcp = pd.DataFrame(df_mcp, columns=['Streams_Mcp_kW/K']).sort_index()

        dict = {}
        for a, b in ampl.getSet('StreamsOfUnit').instances():
            dict[a] = b.getValues().toPandas().index.to_list()
        df_dict = pd.DataFrame.from_dict(dict, orient='index')

        df_Tin = get_ampl_data(ampl, 'Streams_Tin', multi_index=True)
        df_Tout = get_ampl_data(ampl, 'Streams_Tout', multi_index=True)

        df_Tin.index.set_names(['Stream', 'Period', 'Time'], inplace=True)
        df_Tout.index.set_names(['Stream', 'Period', 'Time'], inplace=True)

        df_Tmin = get_ampl_data(ampl, 'dTmin')
        df_Tmin.index.set_names('Stream', inplace=True)

        df_Q = df_Q.reset_index('Service', drop=False)
        df_Q = df_Q.join([df_Tin, df_Tout]).set_index('Service', append=True)
        df_Q = df_Q.merge(df_mcp, left_index=True, right_on=['Stream', 'Period', 'Time', 'Service'])
        df_Q = df_Q.reorder_levels([3, 0, 1, 2]).sort_index()

        df_dict = df_dict.unstack()
        df_dict.dropna(inplace=True)
        df_dict = df_dict.reset_index().set_index(0)
        df_dict.drop(columns='level_0', inplace=True)
        df_dict.index.set_names(['Stream'], inplace=True)
        df_dict.rename(columns={'level_1': 'Unit'}, inplace=True)

        df_Streams_t = df_Q.reset_index().set_index('Stream')
        df_Streams_t = df_Streams_t.join([df_dict, df_Tmin]).set_index(['Service', 'Unit', 'Period', 'Time'], append=True)

        return df_Streams_t.sort_index()

    def set_dfs_lca(ampl):

        LCA_units = get_ampl_data(ampl, 'lca_units')
        LCA_units = LCA_units.stack().unstack(level=0).droplevel(level=1)

        LCA_tot = get_ampl_data(ampl, 'lca_tot')
        LCA_tot = LCA_tot.stack().unstack(level=0)
        LCA_tot.index = ["Network"]
        LCA_tot_house = get_ampl_data(ampl, 'lca_tot_house')
        LCA_tot_house = LCA_tot_house.stack().unstack(level=0).droplevel(1)
        LCA_tot = pd.concat([LCA_tot_house, LCA_tot], axis=0)
        LCA_tot.index.names = ['Hub']

        LCA_op = get_ampl_data(ampl, 'lca_op')
        LCA_op = LCA_op.stack().unstack(level=0).droplevel(level=1)

        return LCA_units, LCA_tot, LCA_op

    def set_dfs_pv(ampl):

        # PV_Surface
        df_PV_Surface = get_ampl_data(ampl, "HouseSurfaceArea", multi_index=True)
        df_PV_Surface.index.names = ['Hub', 'Surface']
        df_PV_Surface.rename(columns={'HouseSurfaceArea': 'Area'})
        df_PV_Surface = df_PV_Surface.sort_index()

        # ["df_PV_orientation"]
        df_PVA_module_nbr = get_ampl_data(ampl, 'PVA_module_nbr', multi_index=True)
        df_PVA_module_nbr = df_PVA_module_nbr.droplevel(4)
        df_PVA_module_nbr.index.names = ['Hub', 'Surface', 'Azimuth', 'Tilt']
        # int_index = df_PVA_module_nbr.index.get_level_values('Surface').astype(int)
        # df_PVA_module_nbr.index.set_levels(int_index, level="Surface", inplace=True)

        df_PVA_module_coverage = get_ampl_data(ampl, "PVA_module_coverage", multi_index=True)
        df_PVA_module_coverage = df_PVA_module_coverage.droplevel(1)
        df_PVA_module_coverage.index.names = ['Hub', 'Surface', 'Azimuth', 'Tilt']

        pd.concat([df_PVA_module_coverage, df_PVA_module_nbr], axis=1)

        df_unshaded_share = get_ampl_data(ampl, 'unshaded_share', multi_index=True)
        df_unshaded_share = df_unshaded_share.groupby(level=[0, 1, 2, 3]).sum()
        df_unshaded_share = df_unshaded_share.div(145)  # 145 patches in skydome
        df_unshaded_share.index.names = ['Hub', 'Surface', 'Azimuth', 'Tilt']

        df_PV_orientation = pd.concat([df_PVA_module_coverage, df_unshaded_share, df_PVA_module_nbr], axis=1)
        # df_PV_Surface_profiles
        # centralized optimization PV_electricity takes a lot of memory -> TODO move to annuals profiles are not needed
        # df_PV_Surface_profiles = get_ampl_data(ampl, "PV_electricity")
        # df_PV_Surface_loss = get_ampl_data(ampl, "PV_electricity_without_loss")
        # df_PV_Surface_profiles = pd.concat([df_PV_Surface_profiles, df_PV_Surface_loss], axis=1)
        # df_PV_Surface_profiles.index.names = ['Hub','Unit','Surface', 'Azimuth','Tilt', 'Period', 'Time']
        # df_Results["df_PV_Surface"]_profiles = df_PV_Surface_profiles.sort_index()

        return df_PV_Surface, df_PV_orientation

    def set_dfs_other(ampl):
        # Time
        df1 = get_ampl_data(ampl, 'dp')
        df2 = get_ampl_data(ampl, 'TimeEnd')
        df3 = get_ampl_data(ampl, 'dt')
        df_Time = pd.concat([df1, df2, df3], axis=1)
        df_Time.index.names = ['Period']
        df_Time = df_Time.sort_index()

        # External
        df1 = get_ampl_data(ampl, 'T_ext', multi_index=True)
        df2 = get_ampl_data(ampl, 'I_global', multi_index=True)
        df_Weather = pd.concat([df1, df2], axis=1)
        df_Weather.index.names = ['Period', 'Time']
        df_Weather = df_Weather.sort_index()

        # Index
        df_Index = get_ampl_data(ampl, 'PeriodOfYear')
        df_Index.index.names = ['HourOfYear']
        df_Index = df_Index.sort_index()

        return df_Time, df_Weather, df_Index

    df_Results = dict()
    df_Results["df_Performance"] = set_df_performance(ampl, scenario)
    df_Results["df_Annuals"] = set_df_annuals(ampl)
    df_Results["df_Unit"], df_Unit_t = set_df_unit(ampl)
    df_Results["df_Grid_t"] = set_df_grid(ampl, method)
    df_Results["df_Time"], df_Weather, df_Index = set_dfs_other(ampl)

    if method['save_data_input']:
        df_Results["df_Buildings"] = set_df_buildings(buildings_data)
        df_Results["df_Weather"] = df_Weather
        df_Results["df_Index"] = df_Index

    if method["save_timeseries"]:
        df_Results["df_Buildings_t"] = set_df_buildings_t(ampl)
        df_Results["df_Unit_t"] = df_Unit_t
    else:
        for i in ['Cost_demand', 'Cost_supply', 'GWP_demand', 'GWP_supply', "Uncontrollable_load"]:
            del df_Results["df_Grid_t"][i]

    if method["save_streams"]:
        df_Results["df_Streams_t"] = set_df_streams_t(ampl)

    if method['save_lca']:
        df_Results["df_lca_Units"], df_Results["df_lca_Performance"], df_Results["df_lca_operation"] = set_dfs_lca(ampl)
    if method['use_pv_orientation'] or method['use_facades']:
        df_Results["df_PV_Surface"], df_Results["df_PV_orientation"] = set_dfs_pv(ampl)
    if method["extract_parameters"]:
        parameters_record = {}
        for p, ampl_obj in ampl.getParameters():
            try:
                parameters_record[p] = ampl.getData(p).toPandas()
            except:
                logging.info(p)

    if filter:
        for df_name, df in df_Results.items():
            df = df.fillna(0)  # replace all NaN with zeros
            df = df.loc[~(df == 0).all(axis=1)]  # drop all lines with only zeros

    return df_Results


def get_df_Results_from_MP(ampl, binary=False, method=None, district=None, read_DHN=False, scenario={}):
    df_Results = dict()

    # Dantzig Wolfe algorithm
    df1 = get_ampl_data(ampl, 'lambda', multi_index=True)
    df_DW = pd.concat([df1], axis=1)
    df_DW.index.names = ['FeasibleSolution', 'Hub']
    df_Results["df_DW"] = df_DW.sort_index()

    if method["save_data_input"] or binary:
        # Building_t
        df1 = get_ampl_data(ampl, 'Grid_supply', multi_index=True)
        df2 = get_ampl_data(ampl, 'Grid_demand', multi_index=True)
        df_Buildings_t = pd.concat([df1, df2], axis=1)
        if read_DHN:
            df3 = get_ampl_data(ampl, 'flowrate_out', multi_index=True)
            df3 = pd.concat({'Heat': df3})
            df4 = get_ampl_data(ampl, 'flowrate_in', multi_index=True)
            df4 = pd.concat({'Heat': df4})
            df_Buildings_t = pd.concat([df_Buildings_t, df3, df4], axis=1)
        df_Buildings_t.index.names = ['Layer', 'FeasibleSolution', 'Hub', 'Period', 'Time']
        df_Results["df_Buildings_t"] = df_Buildings_t.sort_index()

        # Building
        df1 = get_ampl_data(ampl, 'Costs_inv_rep_SPs', multi_index=True)
        df2 = get_ampl_data(ampl, 'Costs_ft_SPs', multi_index=True)
        df_Buildings = pd.concat([df1, df2], axis=1)
        df_Buildings.index.names = ['FeasibleSolution', 'Hub']
        df_Results["df_Buildings"] = df_Buildings.sort_index()

    # District
    df1 = get_ampl_data(ampl, 'Costs_House_op')
    df2 = get_ampl_data(ampl, 'Costs_House_inv')
    df3 = get_ampl_data(ampl, 'Costs_House_cft')
    df4 = df1.values + df2
    df4.columns = ["Costs_House_tot"]
    df5 = get_ampl_data(ampl, 'GWP_House_op')
    df6 = get_ampl_data(ampl, 'GWP_House_constr')
    df_House = pd.concat([df1, df2, df3, df4, df5, df6], axis=1)
    df_House.columns = ["Costs_op", "Costs_inv", "Costs_cft", "Costs_tot", "GWP_op", "GWP_constr"]
    if read_DHN:
        df7 = get_ampl_data(ampl, 'diameter_max')
        df8 = get_ampl_data(ampl, 'DHN_inv_house')
        df8.columns = ["DHN_inv"]
        df9 = get_ampl_data(ampl, 'flowrate_max')
        df_House = pd.concat([df_House, df7, df8, df9], axis=1)

    df1 = get_ampl_data(ampl, 'Costs_op')
    df2 = get_ampl_data(ampl, 'Costs_inv')  # with comfort costs
    df3 = get_ampl_data(ampl, 'Costs_cft')  # with comfort costs
    df4 = get_ampl_data(ampl, 'Costs_tot')  # with comfort costs
    df5 = get_ampl_data(ampl, 'GWP_op')
    df6 = get_ampl_data(ampl, 'GWP_constr')
    df_District = pd.concat([df1, df2, df3, df4, df5, df6], axis=1)
    if read_DHN:
        df7 = np.sqrt(np.sum(df_House[["diameter_max"]] ** 2)).to_frame().transpose()
        df8 = get_ampl_data(ampl, 'DHN_inv')
        df_District = pd.concat([df_District, df7, df8], axis=1)
    df_District = df_District.set_index(pd.Index(["Network"]))
    df_District = pd.concat([df_House, df_District], axis=0)
    df_District.index.names = ['Hub']
    df_Results["df_District"] = df_District.sort_index()

    # df_beta
    emoo_keys = ["EMOO_CAPEX", "EMOO_OPEX", "EMOO_GWP", "EMOO_TOTEX", "EMOO_lca"]
    list_keys = [i for i in scenario["EMOO"].keys() if i in emoo_keys]
    if not list_keys:
        df = pd.DataFrame([0.0] * 16)
        df.index = ['CAPEX', 'OPEX', 'GWP', 'TOTEX'] + list(get_ampl_data(ampl, 'Lca_kpi').index)
        df.columns = ["beta"]
        df_Results["df_beta"] = df
    else:
        df1 = get_ampl_dual_values_in_pandas(ampl, 'EMOO_CAPEX_constraint', False)
        df1.columns = ['CAPEX']
        df2 = get_ampl_dual_values_in_pandas(ampl, 'EMOO_OPEX_constraint', False)
        df2.columns = ['OPEX']
        df3 = get_ampl_dual_values_in_pandas(ampl, 'EMOO_GWP_constraint', False)
        df3.columns = ['GWP']
        df4 = get_ampl_dual_values_in_pandas(ampl, 'EMOO_TOTEX_constraint', False)
        df4.columns = ['TOTEX']
        df_beta = pd.concat([df1, df2, df3, df4], axis=1).stack().droplevel(0)
        df_beta = pd.DataFrame(df_beta, columns=['beta'])
        df5 = get_ampl_dual_values_in_pandas(ampl, 'EMOO_lca_constraint', False)
        df5.columns = ['beta']
        df_Results["df_beta"] = pd.concat([df_beta, df5])

    # District_t
    df1 = get_ampl_data(ampl, 'Cost_demand_network', multi_index=True)
    df2 = get_ampl_data(ampl, 'Cost_supply_network', multi_index=True)
    df3 = get_ampl_data(ampl, 'GWP_demand', multi_index=True)
    df4 = get_ampl_data(ampl, 'GWP_supply', multi_index=True)
    df5 = get_ampl_data(ampl, 'Network_supply', multi_index=True)
    df6 = get_ampl_data(ampl, 'Network_demand', multi_index=True)

    if binary:
        df_District_t = pd.concat([df1, df2, df3, df4, df5, df6], axis=1).sort_index()
    else:
        df_District_t = pd.concat([df5, df6], axis=1)
    df_District_t.index.names = ['Layer', 'Period', 'Time']
    df_Results["df_District_t"] = df_District_t.sort_index()

    # df_Dual
    df1 = get_ampl_dual_values_in_pandas(ampl, 'convexity_1', False)
    df1.columns = ['mu']
    df_Dual = pd.concat([df1], axis=1)
    df_Dual.index.names = ['Hub']
    df_Results["df_Dual"] = df_Dual.sort_index()

    # df_Dual_t
    df1 = get_ampl_dual_values_in_pandas(ampl, 'complicating_cst', True)
    df1.columns = ['pi']
    df2 = get_ampl_dual_values_in_pandas(ampl, 'complicating_cst_GWP', True)
    df2.columns = ['pi_GWP']
    df3 = get_ampl_dual_values_in_pandas(ampl, 'complicating_cst_lca', True).stack().unstack(0).droplevel(3)
    df4 = get_ampl_dual_values_in_pandas(ampl, 'EMOO_grid_constraint', False)
    df4.columns = ['gamma_supply']
    df_Dual_t = pd.concat([df1, df2, df3, df4], axis=1)
    df_Dual_t.index.names = ['Layer', 'Period', 'Time']
    df_Results["df_Dual_t"] = df_Dual_t.sort_index()

    # Unit
    tau = ampl.getParameter('tau').getValues().toList()  # annuality factor
    df1 = get_ampl_data(ampl, 'Units_Use')
    df2 = get_ampl_data(ampl, 'Units_Mult')
    df3 = tau[0] * get_ampl_data(ampl, 'Costs_Unit_inv')
    df4 = get_ampl_data(ampl, 'GWP_Unit_constr')  # per year! For total - multiply with lifetime
    df5 = get_ampl_data(ampl, 'lifetime')
    df_Unit = pd.concat([df1, df2, df3, df4, df5], axis=1)
    if read_DHN:
        df_DHN = pd.DataFrame([[1, 1, get_ampl_data(ampl, 'DHN_inv')["DHN_inv"][0], 0, 0]], index=["DHN"], columns=df_Unit.columns)
        df_Unit = pd.concat([df_Unit, df_DHN], axis=0)
    df_Unit.index.names = ['Unit']
    df_Results["df_Unit"] = df_Unit.sort_index()

    # Unit_t
    df1 = get_ampl_data(ampl, 'Units_demand', multi_index=True)
    df2 = get_ampl_data(ampl, 'Units_supply', multi_index=True)
    df3 = get_ampl_data(ampl, 'BAT_E_stored', multi_index=True)
    df3 = pd.concat([df3], keys=['Electricity'], names=['Layer'])
    df_Unit_t = pd.concat([df1, df2, df3], axis=1)

    if len(district.UnitsOfDistrict) > 0:
        if "EV_district" in district.UnitsOfDistrict:
            df4 = get_ampl_data(ampl, 'EV_E_stored', multi_index=True)
            df4 = pd.concat([df4], keys=['Electricity'], names=['Layer'])
            df5 = get_ampl_data(ampl, 'EV_displacement', multi_index=True)
            df5 = pd.concat([df5], keys=['Electricity'], names=['Layer'])
            df6 = get_ampl_data(ampl, 'EV_V2V', multi_index=True)
            df6 = pd.concat([df6], keys=['Electricity'], names=['Layer'])
            df_Unit_t = pd.concat([df_Unit_t, df4, df5, df6], axis=1)
    df_Unit_t.index.names = ['Layer', 'Unit', 'Period', 'Time']

    units_districts = district.UnitsOfDistrict
    district_l_u = []
    for layer, units in district.UnitsOfLayer.items():
        [district_l_u.append((layer, unit)) for unit in units if unit in units_districts]
    df_Unit_t = df_Unit_t.reset_index(level=['Period', 'Time']).loc[district_l_u, :]
    df_Results["df_Unit_t"] = df_Unit_t.reset_index().set_index(['Layer', 'Unit', 'Period', 'Time']).sort_index()

    # df_lca
    if method["save_lca"]:
        LCA_units = get_ampl_data(ampl, 'lca_units', multi_index=True)
        LCA_units = LCA_units.stack().unstack(level=0).droplevel(level=1)
        df_Results["df_lca_Units"] = LCA_units

        LCA_tot = get_ampl_data(ampl, 'lca_tot')
        LCA_tot = LCA_tot.stack().unstack(level=0)
        LCA_tot.index = ["Network"]
        LCA_tot_house = get_ampl_data(ampl, 'lca_tot_house', multi_index=True)
        LCA_tot_house = LCA_tot_house.stack().unstack(level=0).droplevel(1)
        df_Results["df_lca_Performance"] = pd.concat([LCA_tot_house, LCA_tot], axis=0)
        df_Results["df_lca_Performance"].index.names = ['Hub']

        LCA_op = get_ampl_data(ampl, 'lca_op', multi_index=True)
        LCA_op = LCA_op.stack().unstack(level=0).droplevel(level=1)
        df_Results["df_lca_operation"] = LCA_op

    if method["actors_problem"]:
        df1 = get_ampl_data(ampl, 'Cost_demand_district', multi_index=True).groupby(level=(0, 2)).sum()
        df2 = get_ampl_data(ampl, 'Cost_supply_district', multi_index=True).groupby(level=(0, 2)).sum()
        df3 = get_ampl_data(ampl, 'Cost_self_consumption', multi_index=True).groupby(level=1).sum()
        df3 = pd.concat({'Electricity': df3})
        df_Results["df_Actors_tariff"] = pd.concat([df1, df2, df3], axis=1)

        df_Results["df_Actors"] = get_ampl_data(ampl, 'objective_functions')

        df1 = get_ampl_data(ampl, 'C_op_renters_to_utility')
        df2 = get_ampl_data(ampl, 'C_op_renters_to_owners')
        df3 = get_ampl_data(ampl, 'C_op_utility_to_owners')
        df4 = get_ampl_data(ampl, 'Costs_House_inv')
        df4.columns = ["owner_inv"]
        df5 = get_ampl_data(ampl, 'owner_portfolio')
        df_Actors = pd.concat([df1, df2, df3, df4, df5], axis=1)
        df_6 = df_Actors.sum(axis=0).to_frame().T.set_index(pd.Index(["Network"]))
        df_Actors = pd.concat([df_Actors, df_6], axis=0)
        df_Results["df_District"] = pd.concat([df_Results["df_District"], df_Actors], axis=1)
        df_Results["df_District"].loc["Network", "Objective"] = ampl.getObjective("TOTEX_bui").getValues().toList()[0]

    return df_Results


def get_ampl_data(ampl, ampl_name, multi_index=False):
    # AMPl data in AMPLPY Dataframe
    df = ampl.getData(ampl_name)
    # transform to Pandas Dataframe
    df = df.toPandas()
    # Change index from tuple to multi index
    if multi_index:
        df.index = pd.MultiIndex.from_tuples(df.index)

    return df


def get_ampl_dual_values_in_pandas(ampl, ampl_name, multi_index):
    # AMPl data in AMPLPY Dataframe
    df = ampl.getConstraint(ampl_name).getValues().toPandas()
    # Change index from tuple to multi index
    if multi_index:
        df.index = pd.MultiIndex.from_tuples(df.index)

    return df



import itertools as itertools
import logging

from amplpy import AMPL, Environment

import reho.model.preprocessing.EV_profile_generator as EV_gen
import reho.model.preprocessing.buildings_profiles as buildings_profiles
import reho.model.preprocessing.emissions_parser as emissions
import reho.model.preprocessing.weather as weather
from reho.model.preprocessing.QBuildings import *


__doc__ = """
File for handling data and optimization for an AMPL sub-problem.
"""


class SubProblem:
    """
    Collects all the data input and sends it an AMPL model, solves the optimization.

    Parameters
    ----------
    district : district
        Instance of the class district, contains relevant structure in the district such as Units or grids.
    buildings_data : dict
        Building-specific data.
    local_data : dict
        Location-specific data.
    parameters : dict, optional
        Dictionary containing 'new' parameters for the AMPL model. If incomplete, uses data from buildings_data.
    set_indexed : dict, optional
        Dictionary containing new data which are indexed sets in the AMPL model.
    cluster : dict, optional
        Dictionary containing information about clustering.
    scenario : dict, optional
        Dictionary containing the objective function, EMOO constraints, and additional constraints.
    method : dict, optional
        Dictionary containing different options for methodology choices.
    solver : str, optional
        Chosen solver for AMPL (gurobi, cplex, HiGHS, cbc...).
    qbuildings_data : dict, optional
        Input data for the buildings.

    See also
    --------
    reho.model.reho.reho
    reho.model.master_problem.MasterProblem

    """

    def __init__(self, district, buildings_data, local_data, parameters, set_indexed, cluster, scenario, method, solver, qbuildings_data=None):

        self.buildings_data_sp = buildings_data
        if method['use_facades']:
            self.facades_sp = qbuildings_data['facades_data']
            self.shadows_sp = qbuildings_data['shadows_data']
        if method['use_pv_orientation']:
            self.roofs_sp = qbuildings_data['roofs_data']
        self.infrastructure_sp = district
        self.local_data = local_data
        self.parameters_sp = parameters
        self.set_indexed_sp = set_indexed
        self.cluster_sp = cluster
        if 'exclude_units' not in scenario:
            scenario['exclude_units'] = []
        if 'enforce_units' not in scenario:
            scenario['enforce_units'] = []
        self.scenario_sp = scenario
        self.method_sp = method
        self.solver = solver
        self.parameters_to_ampl = dict()

    def build_model_without_solving(self):
        File_ID = weather.get_cluster_file_ID(self.cluster_sp)

        self.initialize_parameters_for_ampl_and_python()
        ampl = self.init_ampl_model()
        ampl = self.set_weather_data(ampl)
        ampl = self.set_ampl_sets(ampl)
        self.set_emissions_profiles(File_ID)
        self.set_temperature_and_EVs_profiles()
        self.set_HP_parameters(ampl)
        self.set_streams_temperature(ampl)
        if self.method_sp['use_pv_orientation']:
            self.set_PV_models(ampl)
        ampl = self.send_parameters_and_sets_to_ampl(ampl)
        ampl = self.set_scenario(ampl)

        return ampl

    def initialize_parameters_for_ampl_and_python(self):
        # -----------------------------------------------------------------------------------------------------#
        # Default methods
        # -----------------------------------------------------------------------------------------------------#
        self.method_sp = initialize_default_methods(self.method_sp)
        # -----------------------------------------------------------------------------------------------------#
        # Input Parameter Preparation
        # -----------------------------------------------------------------------------------------------------#

        # prepare input parameter which are used in AMPL model

        buildings_to_ampl = ['ERA', 'SolarRoofArea', 'U_h', 'HeatCapacity',
                             'T_comfort_min_0', 'Th_supply_0', 'Th_return_0', 'Tc_supply_0', 'Tc_return_0']

        for parameter in buildings_to_ampl:
            self.parameters_to_ampl[parameter] = {}
            for i, b in enumerate(self.buildings_data_sp):
                if parameter not in self.parameters_sp:
                    self.parameters_to_ampl[parameter][b] = self.buildings_data_sp[b][parameter]
                else:
                    self.parameters_to_ampl[parameter][b] = self.parameters_sp[parameter]

    def init_ampl_model(self):
        if os.getenv('USE_AMPL_MODULES', False):
            from amplpy import modules
            modules.load()
            ampl = AMPL()
        else:
            try:
                ampl = AMPL(Environment(os.environ["AMPL_PATH"]))
            except:
                raise Exception("AMPL_PATH is not defined. Please include a .env file at the project root (e.g., AMPL_PATH='C:/AMPL')")

        # -AMPL (GNU) OPTIONS
        ampl.setOption('solution_round', 11)

        ampl.setOption('presolve_eps', 1e-4)  # -ignore difference between upper and lower bound by this tolerance
        ampl.setOption('presolve_inteps', 1e-6)  # -tolerance added/substracted to each upper/lower bound
        ampl.setOption('presolve_fixeps', 1e-9)
        if not self.method_sp['print_logs']:
            ampl.setOption('show_stats', 0)
            ampl.setOption('solver_msg', 0)

        # -SOLVER OPTIONS
        ampl.setOption('solver', self.solver)
        if self.solver == "gurobi":
            ampl.eval("option gurobi_options 'NodeFileStart=0.5';")
        if self.solver == "cplex":
            ampl.eval("option cplex_options 'bestbound mipgap=5e-7 integrality=1e-09 timing=1 timelimit=3000';")

        # -----------------------------------------------------------------------------------------------------#
        #  MODEL FILES
        # -----------------------------------------------------------------------------------------------------#
        ampl.cd(path_to_ampl_model)
        ampl.read('sub_problem.mod')

        # Energy conversion Units
        ampl.cd(path_to_units)
        if 'ElectricalHeater' in self.infrastructure_sp.UnitTypes:
            ampl.read('electrical_heater.mod')
        if 'NG_Boiler' in self.infrastructure_sp.UnitTypes:
            ampl.read('ng_boiler.mod')
        if 'OIL_Boiler' in self.infrastructure_sp.UnitTypes:
            ampl.read('oil_boiler.mod')
        if 'WOOD_Stove' in self.infrastructure_sp.UnitTypes:
            ampl.read('wood_stove.mod')
        if 'HeatPump' in self.infrastructure_sp.UnitTypes:
            ampl.read('heatpump.mod')
        if 'Air_Conditioner' in self.infrastructure_sp.UnitTypes:
            ampl.read('air_conditioner.mod')
        if 'ThermalSolar' in self.infrastructure_sp.UnitTypes:
            ampl.read('thermal_solar.mod')
        if 'DataHeat' in self.infrastructure_sp.UnitTypes:
            ampl.read('data_heat.mod')
        if 'NG_Cogeneration' in self.infrastructure_sp.UnitTypes:
            ampl.read('ng_cogeneration.mod')
        if 'DHN_hex' in self.infrastructure_sp.UnitTypes:
            ampl.read('DHN_HEX.mod')
            ampl.read('DHN_pipes.mod')
        if 'PV' in self.infrastructure_sp.UnitTypes:
            if self.method_sp['use_pv_orientation']:
                ampl.read('pv_orientation.mod')
            else:
                ampl.read('pv.mod')

        # district Units
        if 'EV' in self.infrastructure_sp.UnitTypes:
            ampl.cd(path_to_district_units)
            ampl.read('evehicle.mod')
        # Storage Units
        ampl.cd(path_to_units_storage)
        if 'WaterTankSH' in self.infrastructure_sp.UnitTypes:
            ampl.read('heatstorage.mod')
        if 'WaterTankDHW' in self.infrastructure_sp.UnitTypes:
            ampl.read('dhwstorage.mod')
        if 'Battery' in self.infrastructure_sp.UnitTypes:
            ampl.read('battery.mod')
        ampl.cd(path_to_ampl_model)

        # Objectives, epsilon constraints and specific constraints
        ampl.read('scenario.mod')

        # TODO: integrate Jules units into district structure (avoid using ampl eval)
        if self.method_sp['use_Storage_Interperiod']:
            ampl.eval(
                'set UnitsOfStorage := setof{u in UnitsOfType["Battery_interperiod"] union UnitsOfType["PTES_storage"]'
                'union UnitsOfType["PTES_conversion"] union UnitsOfType["CH4storage"]'
                'union UnitsOfType["H2storage"] union UnitsOfType["SOEFC"]'
                'union UnitsOfType["Methanizer"] union UnitsOfType["FuelCell"]'
                'union UnitsOfType["Electrolyzer"] union UnitsOfType["WaterTankSH_interperiod"]'
                'union UnitsOfType["SolidLiquidLHS"]'
                '} u;')

            # Storage Units
            ampl.cd(path_to_units_storage)
            ampl.read('h2_storage.mod')
            ampl.read('heatstorage_interperiod.mod')
            ampl.read('LHS_storage.mod')
            ampl.read('battery_interperiod.mod')
            # ampl.read('PTES_aggregated.mod')
            ampl.read('PTES_split.mod')
            ampl.read('CH4_tank.mod')

            # H2 Units
            ampl.cd(path_to_units_h2)
            ampl.read('fuel_cell.mod')
            ampl.read('electrolyser.mod')
            ampl.read('SOEFC.mod')
            ampl.read('methanizer.mod')

            ampl.cd(path_to_units)
            ampl.read('heat_curtailment.mod')
            ampl.cd(path_to_ampl_model)

        return ampl

    def set_weather_data(self, ampl):
        # -----------------------------------------------------------------------------------------------------#
        # -Setting DATA
        # -----------------------------------------------------------------------------------------------------#

        ampl.cd(path_to_clustering)

        File_ID = weather.get_cluster_file_ID(self.cluster_sp)

        ampl.readData('frequency_' + File_ID + '.dat')
        ampl.readData('index_' + File_ID + '.dat')
        self.parameters_to_ampl['T_ext'] = np.loadtxt(os.path.join(path_to_clustering, 'T_' + File_ID + '.dat'))
        self.parameters_to_ampl['I_global'] = np.loadtxt(os.path.join(path_to_clustering, 'Irr_' + File_ID + '.dat'))

        ampl.cd(path_to_ampl_model)

        return ampl

    def set_ampl_sets(self, ampl):
        # -----------------------------------------------------------------------------------------------------#
        # Design Structure: Building Cluster, Units and Layers
        # -----------------------------------------------------------------------------------------------------#

        if self.method_sp['use_discrete_units']:
            self.infrastructure_sp.set_discretize_unit_size()

        self.parameters_to_ampl['Units_flowrate'] = self.infrastructure_sp.Units_flowrate
        self.parameters_to_ampl['Grids_flowrate'] = self.infrastructure_sp.Grids_flowrate
        self.parameters_to_ampl['Grids_Parameters'] = self.infrastructure_sp.Grids_Parameters
        self.parameters_to_ampl['Grids_Parameters_lca'] = self.infrastructure_sp.Grids_Parameters_lca
        self.parameters_to_ampl['Units_Parameters'] = self.infrastructure_sp.Units_Parameters
        self.parameters_to_ampl['Units_Parameters_lca'] = self.infrastructure_sp.Units_Parameters_lca
        self.parameters_to_ampl['Streams_H'] = self.infrastructure_sp.Streams_H

        for key in self.infrastructure_sp.HP_parameters:
            self.parameters_to_ampl[key] = self.infrastructure_sp.HP_parameters[key]

        for s in self.infrastructure_sp.Set:
            if isinstance(self.infrastructure_sp.Set[s], np.ndarray):
                ampl.getSet(str(s)).setValues(self.infrastructure_sp.Set[s])
            elif isinstance(self.infrastructure_sp.Set[s], dict):
                for i, instance in ampl.getSet(str(s)):
                    instance.setValues(self.infrastructure_sp.Set[s][i])
            else:
                raise ValueError('Type Error setting AMPLPY Set', s)

        all_units = [unit for unit, value in ampl.getVariable('Units_Use').instances()]
        for i in all_units:
            for u in self.scenario_sp['exclude_units']:
                if 'district' not in i and u in i:  # unit at the building scale
                    ampl.getVariable('Units_Use').get(str(i)).fix(0)
                elif u in all_units:  # unit at the district scale with problem definition at the district scale
                    ampl.getVariable('Units_Use').get(str(u)).fix(0)

            for u in self.scenario_sp['enforce_units']:
                if 'district' not in i and u in i:  # unit at the building scale
                    ampl.getVariable('Units_Use').get(str(i)).fix(1)  # !!Fmin = 0, leaves the option to exclude unit
                elif u in all_units:  # unit at the district scale with problem definition at the district scale
                    ampl.getVariable('Units_Use').get(str(u)).fix(1)

        return ampl

    def set_emissions_profiles(self, File_ID):

        df_em = emissions.return_typical_emission_profiles(self.cluster_sp, File_ID, 'GWP100a', self.local_data["df_Timestamp"],
                                                           self.local_data["df_Emissions"])
        if self.method_sp['use_dynamic_emission_profiles']:
            self.parameters_to_ampl['GWP_supply'] = df_em
            self.parameters_to_ampl['GWP_demand'] = df_em.rename(columns={'GWP_supply': 'GWP_demand'})
            self.parameters_to_ampl['Gas_emission'] = self.infrastructure_sp.Grids_Parameters.drop('Electricity').drop(
                columns=['Cost_demand_cst', 'Cost_supply_cst'])

    def set_temperature_and_EVs_profiles(self):

        # Reference temperature
        self.parameters_to_ampl['T_comfort_min'] = buildings_profiles.reference_temperature_profile(self.parameters_to_ampl, self.cluster_sp)

        # Set default EV plug out profile if EVs are allowed
        if "EV_plugged_out" not in self.parameters_to_ampl:
            if len(self.infrastructure_sp.UnitsOfDistrict) != 0:
                if "EV_district" in self.infrastructure_sp.UnitsOfDistrict:
                    self.parameters_to_ampl["EV_plugged_out"], self.parameters_to_ampl["EV_plugging_in"] = EV_gen.generate_EV_plugged_out_profiles_district(
                        self.cluster_sp, self.local_data["df_Timestamp"])

    def set_HP_parameters(self, ampl):
        # --------------- Heat Pump ---------------------------------------------------------------------------#
        # T = 7.5C for Lake, T = 12C for underground, T = 16C for CO2

        df_end = ampl.getParameter('TimeEnd').getValues().toPandas()
        timesteps = int(df_end['TimeEnd'].sum())  # total number of timesteps
        sources = []
        if 'T_source' in self.parameters_sp:
            sources = self.parameters_sp['T_source'].keys()

        T_source = []
        if 'HeatPump' in self.infrastructure_sp.UnitsOfType:
            for unit in self.infrastructure_sp.UnitsOfType['HeatPump']:
                if any([i in unit for i in sources]):  # if T_source defined from script
                    source = list(itertools.compress(sources, [i in unit for i in sources]))[0]
                    T_source = np.concatenate([T_source, np.repeat(self.parameters_sp['T_source'][source], timesteps)])
                elif 'Air' in unit:
                    T_source = np.concatenate([T_source, self.parameters_to_ampl['T_ext']])
                elif 'Lake' in unit:
                    T_source = np.concatenate([T_source, np.repeat(7.5, timesteps)])
                elif 'Geothermal' in unit:
                    T_source = np.concatenate([T_source, np.repeat(8, timesteps)])
                elif 'Anergy' in unit:
                    T_source = np.concatenate([T_source, np.repeat(16, timesteps)])
                elif 'DHN' in unit:
                    if 'T_DHN_supply' and 'T_DHN_return' in self.parameters_sp:
                        T_DHN_mean = (self.parameters_sp["T_DHN_supply"] + self.parameters_sp["T_DHN_return"]) / 2
                    elif 'T_DHN_supply_cst' and 'T_DHN_return_cst' in self.parameters_sp:
                        T_DHN_mean = (self.parameters_sp["T_DHN_supply_cst"] + self.parameters_sp["T_DHN_return_cst"]) / 2
                        T_DHN_mean = np.repeat(T_DHN_mean, timesteps)
                    else:
                        T_DHN_mean = np.repeat(16, timesteps)
                    T_source = np.concatenate([T_source, T_DHN_mean])
                else:
                    raise Exception('HP source undefined')

            self.parameters_to_ampl['T_source'] = T_source
            if 'T_source' in self.parameters_sp:
                del self.parameters_sp["T_source"]

        sources = []
        if 'T_source_cool' in self.parameters_sp:
            sources = self.parameters_sp['T_source_cool'].keys()

        T_source_cool = np.array([])
        if 'Air_Conditioner' in self.infrastructure_sp.UnitsOfType:
            for unit in self.infrastructure_sp.UnitsOfType['Air_Conditioner']:
                # if T_source_cool defined from script
                if any([i in unit for i in sources]):
                    source = list(itertools.compress(sources, [i in unit for i in sources]))[0]
                    T_source_cool = np.concatenate([T_source_cool, np.repeat(self.parameters_sp['T_source_cool'][source], timesteps)])

                elif "DHN" in unit:
                    if 'T_DHN_supply' and 'T_DHN_return' in self.parameters_sp:
                        T_DHN_mean = (self.parameters_sp["T_DHN_supply"] + self.parameters_sp["T_DHN_return"]) / 2
                    elif 'T_DHN_supply_cst' and 'T_DHN_return_cst' in self.parameters_sp:
                        T_DHN_mean = (self.parameters_sp["T_DHN_supply_cst"] + self.parameters_sp["T_DHN_return_cst"]) / 2
                        T_DHN_mean = np.repeat(T_DHN_mean, timesteps)
                    else:
                        T_DHN_mean = np.repeat(16, timesteps)
                    T_source_cool = np.concatenate([T_source_cool, T_DHN_mean])

                elif "Air" in unit:
                    T_source_cool = np.concatenate([T_source_cool, self.parameters_to_ampl['T_ext']])

                else:
                    raise Exception('AC sink undefined')

            self.parameters_to_ampl['T_source_cool'] = T_source_cool
            if 'T_source_cool' in self.parameters_sp:
                del self.parameters_sp["T_source_cool"]

    def set_streams_temperature(self, ampl):

        df_end = ampl.getParameter('TimeEnd').getValues().toPandas()
        timesteps = int(df_end['TimeEnd'].sum())
        df_Streams_T = pd.DataFrame(columns=["Period", "Time", "Streams", "Streams_Tout", "Streams_Tin"])
        df_Streams_T = df_Streams_T.set_index(["Period", "Time", "Streams"])

        index = [[(i, j + 1) for j in list(range(int(df_end["TimeEnd"][i])))] for i in df_end.index]
        index = [j for i in index for j in i]
        index = pd.MultiIndex.from_tuples(index, names=["Period", "Time"])

        for bui in self.infrastructure_sp.houses:
            for unit_data in self.infrastructure_sp.houses[bui]["units"]:
                for i, T_level in enumerate(unit_data["StreamsOfUnit"]):
                    stream = unit_data["name"] + '_' + bui + '_' + T_level
                    df = pd.DataFrame(np.repeat(stream, timesteps), index=index, columns=["Streams"])
                    df["Streams_Tout"] = unit_data["stream_Tout"][i]
                    df["Streams_Tin"] = unit_data["stream_Tin"][i]
                    df.set_index("Streams", append=True, inplace=True)
                    df_Streams_T = pd.concat([df_Streams_T, df])
            for stream in self.infrastructure_sp.StreamsOfBuilding[bui]:
                df = pd.DataFrame(np.repeat(stream, timesteps), index=index, columns=["Streams"])
                df["Streams_Tout"] = 40  # default value that is changed in data_stream.dat
                df["Streams_Tin"] = 50  # default value that is changed in data_stream.dat
                df.set_index("Streams", append=True, inplace=True)
                df_Streams_T = pd.concat([df_Streams_T, df])

        self.parameters_to_ampl['streams_T'] = df_Streams_T.reorder_levels([2, 0, 1])

    def set_PV_models(self, ampl):
        # --------------- PV Panels ---------------------------------------------------------------------------#

        df_dome = skydome.skydome_to_df(self.local_data)
        self.parameters_to_ampl['Sin_a'] = df_dome.Sin_a.values
        self.parameters_to_ampl['Cos_a'] = df_dome.Cos_a.values
        self.parameters_to_ampl['Sin_e'] = df_dome.Sin_e.values
        self.parameters_to_ampl['Cos_e'] = df_dome.Cos_e.values

        df_irr = skydome.irradiation_to_df(ampl, self.local_data["df_Irradiation"], self.local_data["df_Timestamp"])
        self.parameters_to_ampl['Irr'] = df_irr
        # On Flat Roofs optimal Orientation of PV panel is chosen by the solver, Construction of possible Configurations
        # Azimuth = np.array([])
        # Tilt = np.array([])
        Azimuth = np.array(range(160, 210, 10))
        Tilt = np.array([5, 10, 20, 30, 40])

        All_azimuth = np.repeat(Azimuth, len(Tilt))
        All_tilt = np.tile(Tilt, len(Azimuth))

        Configs_flat_roof = [None] * (len(All_azimuth) + len(All_tilt))
        Configs_flat_roof[::2] = All_azimuth
        Configs_flat_roof[1::2] = All_tilt
        Configs_flat_roof.append(180)
        Configs_flat_roof.append(0)

        np_surface = np.array([])
        np_flat_roof = np.array([])
        np_tilted_roof = np.array([])
        dict_SurfaceofHouse = {}
        dict_config = {}
        df_SurfaceArea = pd.DataFrame()
        self.set_indexed_sp['Surface'] = np.array([])

        for b in self.buildings_data_sp:

            df_roofs = self.roofs_sp[self.roofs_sp['id_building'] == self.buildings_data_sp[b]['id_building']]
            # df_profiles is not used, but precalculated by the ampl model
            #  Surface/ Roof Area Values are selected matching to egid
            np_surface = np.append(np_surface, df_roofs['ROOF_ID'].values)
            dict_SurfaceofHouse[b] = df_roofs['ROOF_ID'].values

            # Flat roof if Tilt of Roof is either 1 or 0/ Tilted Roofs are all opposite to flat roofs
            Flat_roofs = df_roofs.ROOF_ID.loc[(df_roofs['TILT'] == 1) | (df_roofs['TILT'] == 0)].values
            Tilted_roofs = df_roofs.ROOF_ID.loc[(df_roofs['TILT'] != 1) & (df_roofs['TILT'] != 0)].values
            np_flat_roof = np.append(np_flat_roof, Flat_roofs)
            np_tilted_roof = np.append(np_tilted_roof, Tilted_roofs)

            # PV Panels are orientated flat on tilted roof --> one configuration possibility
            for tr in Tilted_roofs:
                az = df_roofs.AZIMUTH.loc[(df_roofs['ROOF_ID'] == tr)].values
                ti = df_roofs.TILT.loc[(df_roofs['ROOF_ID'] == tr)].values
                dict_config[tr] = [az[0], ti[0]]

            for fr in Flat_roofs:
                dict_config[fr] = Configs_flat_roof

            # get area of surface
            index = pd.MultiIndex.from_tuples([(b, s) for s in df_roofs['ROOF_ID'].values])
            df = pd.DataFrame(df_roofs['AREA'].values, index=index, columns=['HouseSurfaceArea'])
            df_SurfaceArea = pd.concat([df_SurfaceArea, df])

        self.parameters_to_ampl['HouseSurfaceArea'] = df_SurfaceArea
        self.set_indexed_sp['Surface'] = np.append(self.set_indexed_sp['Surface'], np_surface)
        self.set_indexed_sp['SurfaceOfHouse'] = dict_SurfaceofHouse

        self.set_indexed_sp['SurfaceTypes'] = np.array(['Flat_roof', 'Tilted_roof', 'Facades'])
        self.set_indexed_sp['SurfaceOfType'] = {'Flat_roof': np_flat_roof, 'Tilted_roof': np_tilted_roof, 'Facades': []}
        self.set_indexed_sp['ConfigOfSurface'] = dict_config

        np_facades = np.array([])
        df_limit_angle = pd.DataFrame()

        if self.method_sp['use_facades']:
            for b in self.buildings_data_sp:
                df_facades = self.facades_sp[self.facades_sp['id_building'] == self.buildings_data_sp[b]['id_building']]
                df_shadows = self.shadows_sp[self.shadows_sp['id_building'] == self.buildings_data_sp[b]['id_building']]
                facades = df_facades['Facades_ID']
                np_facades = np.append(np_facades, facades)
                df_shadow = return_shadows_id_building(self.buildings_data_sp[b]['id_building'], df_shadows, self.local_data)
                df_shadow = pd.concat([df_shadow], keys=[b], names=['House'])
                df_limit_angle = pd.concat([df_limit_angle, df_shadow])
                for fc in facades:
                    az = df_facades.AZIMUTH.loc[(df_facades['Facades_ID'] == fc)].values
                    # Tilt is not available for facades
                    # ti = df_facades.TILT.loc[(df_facades['Facades_ID'] == fc)].values
                    ti = 0
                    self.set_indexed_sp['ConfigOfSurface'][fc] = [az[0], ti]
                self.set_indexed_sp['SurfaceOfHouse'][b] = np.append(
                    self.set_indexed_sp['SurfaceOfHouse'][b],
                    df_facades['Facades_ID'].values)
                index = pd.MultiIndex.from_tuples([(b, f) for f in facades])
                df = pd.DataFrame(df_facades['AREA'].values, index=index,
                                  columns=['HouseSurfaceArea'])  # Pq attribuer ici HouseSurfaceArea comme l'aire des facades?
                self.parameters_to_ampl['HouseSurfaceArea'] = pd.concat(
                    [self.parameters_to_ampl['HouseSurfaceArea'], df])
                # self.parameters_to_ampl['HouseSurfaceArea'].sort_index(inplace = True)

            if not df_limit_angle.empty:
                # self.parameters_to_ampl['Limiting_angle_shadow'] = df_limit_angle.rename(columns={0:'Limiting_angle_shadow'})
                self.parameters_to_ampl['Limiting_angle_shadow'] = df_limit_angle
                self.set_indexed_sp['SurfaceOfType']['Facades'] = np_facades
                self.set_indexed_sp['Surface'] = np.append(self.set_indexed_sp['Surface'], np_facades)

    def send_parameters_and_sets_to_ampl(self, ampl):
        # -----------------------------------------------------------------------------------------------------#
        # Load data to AMPLPY depending on their type
        # -----------------------------------------------------------------------------------------------------#

        for key in self.parameters_sp:
            self.parameters_to_ampl[key] = self.parameters_sp[key]

        # set new indexed sets
        for s in self.set_indexed_sp:
            if isinstance(self.set_indexed_sp[s], np.ndarray):
                ampl.getSet(str(s)).setValues(self.set_indexed_sp[s])
            elif isinstance(self.set_indexed_sp[s], dict):
                for i, instance in ampl.getSet(str(s)):
                    try:
                        instance.setValues(self.set_indexed_sp[s][i])
                    except ValueError:
                        instance.setValues([self.set_indexed_sp[s][i]])
            else:
                raise ValueError('Type Error setting AMPLPY Set', s)

        # set new input Parameter
        for i in self.parameters_to_ampl:

            if isinstance(self.parameters_to_ampl[i], np.ndarray):
                Para = ampl.getParameter(i)
                Para.setValues(self.parameters_to_ampl[i])

            elif isinstance(self.parameters_to_ampl[i], list):
                Para = ampl.getParameter(i)
                Para.setValues(np.array(self.parameters_to_ampl[i]))

            elif isinstance(self.parameters_to_ampl[i], pd.DataFrame):
                ampl.setData(self.parameters_to_ampl[i])

            elif isinstance(self.parameters_to_ampl[i], pd.Series):
                self.parameters_to_ampl[i].name = i
                df = pd.DataFrame(self.parameters_to_ampl[i])
                ampl.setData(df)

            elif isinstance(self.parameters_to_ampl[i], dict):
                Para = ampl.getParameter(i)
                Para.setValues(self.parameters_to_ampl[i])

            elif isinstance(self.parameters_to_ampl[i], float):
                Para = ampl.getParameter(i)
                Para.setValues([self.parameters_to_ampl[i]])

            elif isinstance(self.parameters_to_ampl[i], int):
                Para = ampl.getParameter(i)
                Para.setValues([self.parameters_to_ampl[i]])

            else:
                raise ValueError('Type Error setting AMPLPY Parameter', i)

        # TODO remove data_stream.dat
        ampl.readData('data_stream.dat')
        if self.method_sp['use_Storage_Interperiod']:
            ampl.readData('data_stream_storage.dat')

        return ampl

    def set_scenario(self, ampl):

        # Set objective function
        for objective_name, objective_formulation in ampl.getObjectives():
            objective_formulation.drop()

        if 'Objective' in self.scenario_sp:
            try:
                ampl.getObjective(self.scenario_sp['Objective']).restore()
            except KeyError:
                ampl.getObjective('TOTEX').restore()
                logging.warning('Objective function "' + str(self.scenario_sp['Objective']) +
                                '" was not found in ampl model, TOTEX minimization was set instead.')
        else:
            ampl.getObjective('TOTEX').restore()
            logging.warning('No objective function was found in scenario dictionary, TOTEX minimization was set instead.')

        # Set epsilon constraints
        ampl.getConstraint('EMOO_CAPEX_constraint').drop()
        ampl.getConstraint('EMOO_OPEX_constraint').drop()
        ampl.getConstraint('EMOO_TOTEX_constraint').drop()
        ampl.getConstraint('EMOO_GWP_constraint').drop()
        ampl.getConstraint('EMOO_lca_constraint').drop()

        ampl.getConstraint('EMOO_GU_demand_constraint').drop()
        ampl.getConstraint('EMOO_GU_supply_constraint').drop()
        ampl.getConstraint('EMOO_grid_constraint').drop()
        ampl.getConstraint('EMOO_network_constraint').drop()

        if 'EMOO' in self.scenario_sp:
            for epsilon_constraint in self.scenario_sp['EMOO']:
                try:
                    ampl.getConstraint(epsilon_constraint + '_constraint').restore()
                    if isinstance(self.scenario_sp['EMOO'][epsilon_constraint], dict):
                        epsilon_parameter = ampl.getParameter(epsilon_constraint)
                        epsilon_parameter.setValues(self.scenario_sp['EMOO'][epsilon_constraint])
                    else:
                        epsilon_parameter = ampl.getParameter(epsilon_constraint)
                        epsilon_parameter.setValues([self.scenario_sp['EMOO'][epsilon_constraint]])
                except:
                    logging.warning('EMOO constraint ' + str(epsilon_constraint) + ' was not found in ampl model and was thus ignored.')

        # Set specific constraints
        ampl.getConstraint('disallow_exchanges_1').drop()
        ampl.getConstraint('disallow_exchanges_2').drop()
        ampl.getConstraint('no_ElectricalHeater_without_HP').drop()

        if 'PV' in self.infrastructure_sp.UnitsOfType:
            ampl.getConstraint('enforce_PV_max').drop()
        if 'HeatPump' in self.infrastructure_sp.UnitsOfType:
            ampl.getConstraint('enforce_DHN').drop()
            if not any("DHN" in unit for unit in self.infrastructure_sp.UnitsOfType['HeatPump']):
                ampl.getConstraint('DHN_heat').drop()
        if 'Air_Conditioner' in self.infrastructure_sp.UnitsOfType and "Air_Conditioner_DHN" not in [unit["name"] for unit in self.infrastructure_sp.units]:
            ampl.getConstraint('AC_c3').drop()
        if 'EV' in self.infrastructure_sp.UnitTypes:
            ampl.getConstraint('unidirectional_service').drop()

        if self.method_sp['use_pv_orientation']:
            ampl.getConstraint('enforce_PV_max_fac').drop()
            if not self.method_sp['use_facades']:
                ampl.getConstraint('limits_maximal_PV_to_fac').drop()

        if 'specific' in self.scenario_sp:
            for specific_constraint in self.scenario_sp['specific']:
                try:
                    ampl.getConstraint(specific_constraint).restore()
                except:
                    logging.warning('Specific constraint "' + str(specific_constraint) + '" was not found in ampl model and was thus ignored.')

        return ampl

    def solve_model(self):
        ampl = self.build_model_without_solving()

        debugging = False
        if debugging:
            # ampl.exportData('loaded_data.dat')
            # ampl.expotModel('loaded_model.mod')
            ampl.eval('suffix iis symbolic OUT;')
            ampl.setOption('presolve', 1)

        # ampl evaluation to speed up the optimization in case we have cogeneration units (since it has many integer variables)
        # ampl.eval("suffix priority IN, integer, >= 0, <= 9999; let {u in Units} Units_Use[u].priority := 9999; "
        #          "let {u in UnitsOfType['NG_Cogeneration'],p in Period,t in Time[p]} Units_Use_t[u,p,t].priority := "
        #          "200*round(max{i in Period,j in Time[i]}(T_ext[i,j])-T_ext[p,t],0);")
        ampl.solve()

        if debugging:
            ampl.eval('display {i in 1.._ncons: _con[i].iis <> "0"} (_conname[i], _con[i].iis);')
            ampl.eval('for {i in 1.._ncons: _con[i].iis <> "0"} expand _con[i]; ')
            ampl.eval('display{j in 1.._nvars: _var[j].iis <> "0"}(_varname[j], _var[j].iis);')

        exitcode = exitcode_from_ampl(ampl)
        return ampl, exitcode


def initialize_default_methods(method):
    if method is None:
        method = {}

    if 'use_facades' not in method:
        method['use_facades'] = False
    if 'use_pv_orientation' not in method:
        method['use_pv_orientation'] = False

    if 'include_stochasticity' not in method:
        method['include_stochasticity'] = False
    if 'sd_stochasticity' not in method:
        method['sd_stochasticity'] = [0.1, 1]

    if 'building-scale' not in method:
        method['building-scale'] = False
    if 'district-scale' not in method:
        method['district-scale'] = False
    if 'parallel_computation' not in method:
        method['parallel_computation'] = True
    if 'switch_off_second_objective' not in method:
        method['switch_off_second_objective'] = False

    if 'use_discrete_units' not in method:
        method['use_discrete_units'] = False
    if 'fix_units' not in method:
        method['fix_units'] = False

    if 'use_dynamic_emission_profiles' not in method:
        method['use_dynamic_emission_profiles'] = False
    if 'use_custom_profiles' not in method:
        method['use_custom_profiles'] = False

    if 'include_all_solutions' not in method:
        method['include_all_solutions'] = False
    if 'save_data_input' not in method:
        method['save_data_input'] = True
    if 'save_timeseries' not in method:
        method['save_timeseries'] = True
    if 'save_streams' not in method:
        method['save_streams'] = False
    if 'save_lca' not in method:
        method['save_lca'] = False
    if 'extract_parameters' not in method:
        method['extract_parameters'] = False
    if 'print_logs' not in method:
        method['print_logs'] = True

    if 'actors_problem' not in method:
        method['actors_problem'] = False
    if method['actors_problem']:
        method["include_all_solutions"] = True
    if 'DHN_CO2' not in method:
        method['DHN_CO2'] = False

    if 'use_Storage_Interperiod' not in method:
        method['use_Storage_Interperiod'] = False

    if method['building-scale']:
        method['include_all_solutions'] = False  # avoid interactions between optimization scenarios
        method[
            'district-scale'] = True  # building-scale approach is also using the decomposition algorithm, but with only 1 MP optimization (DW_params['max_iter'] = 1)

    return method


def exitcode_from_ampl(ampl):
    solve_result = ampl.getData('solve_result').toList()[0]
    return 0 if solve_result == 'solved' else solve_result
